from typing import Callable, Tuple, Union

import numpy as np

from .ensemble import Ensemble, EnsemblePrediction
from .util import get_prediction_dataset


def get_umap_representation_parameters(
    ens: Ensemble,
    n_components: int = 2,
    normalize_data: bool = False,
    **kwargs,
) -> Tuple:
    """
    UMAP of parameter ensemble.

    Compute the representation with reduced dimensionality via umap
    (with a given number of umap components) of the parameter ensemble.
    Allows to pass on additional keyword arguments to the umap routine.

    Parameters
    ----------
    ens:
        Ensemble objects containing a set of parameter vectors
    n_components:
        number of components for the dimension reduction
    normalize_data:
        flag indicating whether the parameter ensemble should be rescaled with
        mean and standard deviation

    Returns
    -------
    umap_components:
        first components of the umap embedding
    umap_object:
        returned fitted umap object from umap.UMAP()
    """
    # call lowlevel routine using the parameter vector ensemble
    return _get_umap_representation_lowlevel(
        dataset=ens.x_vectors.transpose(),
        n_components=n_components,
        normalize_data=normalize_data,
        **kwargs,
    )


def get_umap_representation_predictions(
    ens: Union[Ensemble, EnsemblePrediction],
    prediction_index: int = 0,
    n_components: int = 2,
    normalize_data: bool = False,
    **kwargs,
) -> Tuple:
    """
    UMAP of ensemble prediction.

    Compute the representation with reduced dimensionality via umap
    (with a given number of umap components) of the ensemble predictions.
    Allows to pass on additional keyword arguments to the umap routine.

    Parameters
    ----------
    ens:
        Ensemble objects containing a set of parameter vectors and a set of
        predictions or EnsemblePrediction object containing only predictions
    prediction_index:
        index telling which prediction from the list should be analyzed
    n_components:
        number of components for the dimension reduction
    normalize_data:
        flag indicating whether the parameter ensemble should be rescaled with
        mean and standard deviation

    Returns
    -------
    umap_components:
        first components of the umap embedding
    umap_object:
        returned fitted umap object from umap.UMAP()
    """
    # extract the an array of predictions from either an Ensemble object or an
    # EnsemblePrediction object
    dataset = get_prediction_dataset(ens, prediction_index)

    # call lowlevel routine using the prediction ensemble
    return _get_umap_representation_lowlevel(
        dataset=dataset,
        n_components=n_components,
        normalize_data=normalize_data,
        **kwargs,
    )


def get_pca_representation_parameters(
    ens: Ensemble,
    n_components: int = 2,
    rescale_data: bool = True,
    rescaler: Union[Callable, None] = None,
) -> Tuple:
    """
    PCA of parameter ensemble.

    Compute the representation with reduced dimensionality via principal
    component analysis (with a given number of principal components) of the
    parameter ensemble.

    Parameters
    ----------
    ens:
        Ensemble objects containing a set of parameter vectors
    n_components:
        number of components for the dimension reduction
    rescale_data:
        flag indicating whether the principal components should be rescaled
        using a rescaler function (e.g., an arcsinh function)
    rescaler:
        callable function to rescale the output of the PCA (defaults to
        numpy.arcsinh)

    Returns
    -------
    principal_components:
        principal components of the parameter vector ensemble
    pca_object:
        returned fitted pca object from sklearn.decomposition.PCA()
    """
    return _get_pca_representation_lowlevel(
        dataset=ens.x_vectors.transpose(),
        n_components=n_components,
        rescale_data=rescale_data,
        rescaler=rescaler,
    )


def get_pca_representation_predictions(
    ens: Union[Ensemble, EnsemblePrediction],
    prediction_index: int = 0,
    n_components: int = 2,
    rescale_data: bool = True,
    rescaler: Union[Callable, None] = None,
) -> Tuple:
    """
    PCA of ensemble prediction.

    Compute the representation with reduced dimensionality via principal
    component analysis (with a given number of principal components) of the
    ensemble prediction.

    Parameters
    ----------
    ens:
        Ensemble objects containing a set of parameter vectors and a set of
        predictions or EnsemblePrediction object containing only predictions
    prediction_index:
        index telling which prediction from the list should be analyzed
    n_components:
        number of components for the dimension reduction
    rescale_data:
        flag indicating whether the principal components should be rescaled
        using a rescaler function (e.g., an arcsinh function)
    rescaler:
        callable function to rescale the output of the PCA (defaults to
        numpy.arcsinh)

    Returns
    -------
    principal_components:
        principal components of the parameter vector ensemble
    pca_object:
        returned fitted pca object from sklearn.decomposition.PCA()
    """
    # extract the an array of predictions from either an Ensemble object or an
    # EnsemblePrediction object
    dataset = get_prediction_dataset(ens, prediction_index)

    # call lowlevel routine using the prediction ensemble
    return _get_pca_representation_lowlevel(
        dataset=dataset,
        n_components=n_components,
        rescale_data=rescale_data,
        rescaler=rescaler,
    )


def _get_umap_representation_lowlevel(
    dataset: np.ndarray,
    n_components: int = 2,
    normalize_data: bool = False,
    **kwargs,
) -> Tuple:
    """
    Low level UMAP of parameter ensemble.

    Compute the representation with reduced dimensionality via uniform
    manifold approximation and projection (with a given number of principal
    components) of the parameter ensemble.

    Parameters
    ----------
    dataset:
        numpy array containing either the ensemble predictions or the parameter
        ensemble itself
    n_components:
        number of components for the dimension reduction
    rescale_data:
        flag indicating whether the principal components should be rescaled
        using a rescaler function (e.g., an arcsinh function)
    rescaler:
        callable function to rescale the output of the PCA (defaults to
        numpy.arcsinh)

    Returns
    -------
    umap_components:
        first components of the umap embedding
    umap_object:
        returned fitted umap object from umap.UMAP()
    """
    import umap
    import umap.plot
    from sklearn.preprocessing import StandardScaler

    # create a umap object
    umap_object = umap.UMAP(n_components=n_components, **kwargs)

    # normalize data with mean and standard deviation if wanted
    if normalize_data:
        dataset = StandardScaler().fit_transform(dataset)

    # perform the manifold fitting and transform the dataset
    umap_components = umap_object.fit_transform(dataset)

    return umap_components, umap_object


def _get_pca_representation_lowlevel(
    dataset: np.ndarray,
    n_components: int = 2,
    rescale_data: bool = True,
    rescaler: Union[Callable, None] = None,
) -> Tuple:
    """
    Low level PCA of parameter ensemble.

    Compute the representation with reduced dimensionality via principal
    component analysis (with a given number of principal components) of the
    parameter ensemble.

    Parameters
    ----------
    dataset:
        numpy array containing either the ensemble predictions or the parameter
        ensemble itself
    n_components:
        number of components for the dimension reduction
    rescale_data:
        flag indicating whether the principal components should be rescaled
        using a rescaler function (e.g., an arcsinh function)
    rescaler:
        callable function to rescale the output of the PCA (defaults to
        numpy.arcsinh)

    Returns
    -------
    principal_components:
        principal components of the parameter vector ensemble
    pca_object:
        returned fitted pca object from sklearn.decomposition.PCA()
    """
    import sklearn.decomposition

    # create a PCA object and decompose the dataset
    pca_object = sklearn.decomposition.PCA(n_components=n_components)
    pca_object.fit(dataset)
    # get the projection down to the first components
    principal_components = pca_object.transform(dataset)

    # rescale the principal components with a non-linear function, if wanted
    if rescale_data:
        if rescaler is None:
            # use arcsinh as default
            principal_components = np.arcsinh(principal_components)
        else:
            # use provided funcation for rescaling
            principal_components = rescaler(principal_components)

    return principal_components, pca_object



# noqa: D400,D205
"""
Sample
======

Draw samples from the distribution, with support for various samplers.
"""

from .adaptive_metropolis import AdaptiveMetropolisSampler
from .adaptive_parallel_tempering import AdaptiveParallelTemperingSampler
from .diagnostics import auto_correlation, effective_sample_size, geweke_test
from .dynesty import DynestySampler
from .emcee import EmceeSampler
from .metropolis import MetropolisSampler
from .parallel_tempering import ParallelTemperingSampler
from .sample import sample
from .sampler import InternalSampler, Sampler
from .util import calculate_ci_mcmc_sample, calculate_ci_mcmc_sample_prediction



"""Manage all components of a pyPESTO model selection problem."""
from typing import Any, Dict, Iterable, List, Optional, Tuple

import petab_select
from petab_select import Model

from ..C import TYPE_POSTPROCESSOR
from .method import MethodCaller


class Problem:
    """Handles use of a model selection algorithm.

    Handles model selection. Usage involves initialisation with a model
    specifications file, and then calling the `select()` method to perform
    model selection with a specified algorithm and criterion.

    Attributes
    ----------
    calibrated_models:
        Storage for all calibrated models. A dictionary, where keys are
        model hashes, and values are `petab_select.Model` objects.
    newly_calibrated_models:
        Storage for models that were calibrated in the previous iteration of
        model selection. Same type as `calibrated_models`.
    method_caller:
        A `MethodCaller`, used to run a single iteration of a model
        selection method.
    model_postprocessor:
        A method that is applied to each model after calibration.
    petab_select_problem:
        A PEtab Select problem.
    """

    # FIXME rename `best_model` to `selected_model` everywhere

    def __init__(
        self,
        petab_select_problem: petab_select.Problem,
        model_postprocessor: Optional[TYPE_POSTPROCESSOR] = None,
    ):
        self.petab_select_problem = petab_select_problem
        self.model_postprocessor = model_postprocessor

        self.set_state(
            calibrated_models={},
            newly_calibrated_models={},
        )

        # TODO default caller, based on petab_select.Problem
        self.method_caller = None

    def create_method_caller(self, **kwargs) -> MethodCaller:
        """Create a method caller.

        `args` and `kwargs` are passed to the `MethodCaller` constructor.

        Returns
        -------
        MethodCaller
            A `MethodCaller` instance.
        """
        return MethodCaller(
            petab_select_problem=self.petab_select_problem,
            calibrated_models=self.calibrated_models,
            model_postprocessor=self.model_postprocessor,
            **kwargs,
        )

    def set_state(
        self,
        calibrated_models: Dict[str, Model],
        newly_calibrated_models: Dict[str, Model],
    ) -> None:
        """Set the state of the problem.

        See :class:`Problem` attributes for argument documentation.
        """
        self.calibrated_models = calibrated_models
        self.newly_calibrated_models = newly_calibrated_models

    def update_with_newly_calibrated_models(
        self,
        newly_calibrated_models: Optional[Dict[str, Model]] = None,
    ) -> None:
        """Update the state of the problem with newly calibrated models.

        Args:
            newly_calibrated_models:
                See attributes of :class:`Problem`.
        """
        self.newly_calibrated_models = newly_calibrated_models
        self.calibrated_models.update(self.newly_calibrated_models)

    def handle_select_kwargs(
        self,
        kwargs: Dict[str, Any],
    ):
        """Check keyword arguments to select calls."""
        if "newly_calibrated_models" in kwargs:
            raise ValueError(
                'Please supply `newly_calibrated_models` via '
                '`pypesto.select.Problem.set_state`.'
            )
        if "calibrated_models" in kwargs:
            raise ValueError(
                'Please supply `calibrated_models` via '
                '`pypesto.select.Problem.set_state`.'
            )

    def select(
        self,
        **kwargs,
    ) -> Tuple[Model, Dict[str, Model], Dict[str, Model]]:
        """Run a single iteration of a model selection algorithm.

        The result is the selected model for the current run, independent of
        previous selected models.

        `kwargs` are passed to the `MethodCaller` constructor.

        Returns
        -------
        tuple
            A 3-tuple, with the following values:

               1. the best model;
               2. all candidate models in this iteration, as a `dict` with
                  model hashes as keys and models as values; and
               3. all candidate models from all iterations, as a `dict` with
                  model hashes as keys and models as values.
        """
        # TODO move some options to PEtab Select? e.g.:
        # - startpoint_latest_mle
        # - select_first_improvement
        self.handle_select_kwargs(kwargs)
        # TODO handle bidirectional
        method_caller = self.create_method_caller(**kwargs)
        previous_best_model, newly_calibrated_models = method_caller(
            # TODO add predecessor model to state
            newly_calibrated_models=self.newly_calibrated_models,
        )

        self.update_with_newly_calibrated_models(
            newly_calibrated_models=newly_calibrated_models,
        )

        best_model = petab_select.ui.best(
            problem=self.petab_select_problem,
            models=self.newly_calibrated_models.values(),
            criterion=method_caller.criterion,
        )

        # TODO: Reconsider return value. `result` could be stored in attribute,
        # then no values need to be returned, and users can request values
        # manually.
        return best_model, newly_calibrated_models

    def select_to_completion(
        self,
        **kwargs,
    ) -> List[Model]:
        """Run an algorithm until an exception `StopIteration` is raised.

        `kwargs` are passed to the `MethodCaller` constructor.

        An exception `StopIteration` is raised by
        `pypesto.select.method.MethodCaller.__call__` when no candidate models
        are found.

        Returns
        -------
        list
            The best models (the best model at each iteration).
        """
        best_models = []
        self.handle_select_kwargs(kwargs)
        method_caller = self.create_method_caller(**kwargs)

        intermediate_kwargs = {}
        while True:
            # TODO currently uses the best model so far, not the best model
            #      from the previous iteration. Make this a possibility?
            # TODO string literals
            if best_models:
                intermediate_kwargs["predecessor_model"] = best_models[-1]
            try:
                previous_best_model, newly_calibrated_models = method_caller(
                    **intermediate_kwargs,
                    newly_calibrated_models=self.newly_calibrated_models,
                )
                self.update_with_newly_calibrated_models(
                    newly_calibrated_models=newly_calibrated_models,
                )
                best_models.append(previous_best_model)
            except StopIteration:
                previous_best_model = (
                    method_caller.candidate_space.predecessor_model
                )
                best_models.append(previous_best_model)
                break

        return best_models

    # TODO method that automatically generates initial models, for a specific
    # number of starts. TODO parallelise?
    def multistart_select(
        self,
        predecessor_models: Iterable[Model] = None,
        **kwargs,
    ) -> Tuple[Model, List[Model]]:
        """Run an algorithm multiple times, with different predecessor models.

        Note that the same method caller is currently shared between all calls.
        This may change when parallelization is implemented, but for now
        ensures that the same model isn't calibrated twice.
        Could also be managed by sharing the same "calibrated_models" object
        (but then the same model could be repeatedly calibrated, if the
        calibrations start before any have stopped).

        `kwargs` are passed to the `MethodCaller` constructor.

        Parameters
        ----------
        predecessor_models:
            The models that will be used as initial models. One "model
            selection iteration" will be run for each predecessor model.

        Returns
        -------
        tuple
            A 2-tuple, with the following values:

               1. the best model; and
               2. the best models (the best model at each iteration).
        """
        self.handle_select_kwargs(kwargs)
        model_lists = []
        newly_calibrated_models_list = [
            self.newly_calibrated_models for _ in predecessor_models
        ]

        method_caller = self.create_method_caller(**kwargs)
        for start_index, predecessor_model in enumerate(predecessor_models):
            (
                best_model,
                newly_calibrated_models_list[start_index],
            ) = method_caller(
                predecessor_model=predecessor_model,
                newly_calibrated_models=newly_calibrated_models_list[
                    start_index
                ],
            )
            self.calibrated_models.update(
                newly_calibrated_models_list[start_index]
            )

            model_lists.append(
                newly_calibrated_models_list[start_index].values()
            )
            method_caller.candidate_space.reset()

        best_model = petab_select.ui.best(
            problem=method_caller.petab_select_problem,
            models=[model for models in model_lists for model in models],
            criterion=method_caller.criterion,
        )

        return best_model, model_lists



"""Profiling result."""

import copy

import numpy as np


class ProfilerResult(dict):
    """
    The result of a profiler run.

    The standardized return value from pypesto.profile, which can
    either be initialized from an OptimizerResult or from an existing
    ProfilerResult (in order to extend the computation).

    Can be used like a dict.

    Attributes
    ----------
    x_path:
        The path of the best found parameters along the profile
        (Dimension: n_par x n_profile_points)
    fval_path:
        The function values, fun(x), along the profile.
    ratio_path:
        The ratio of the posterior function along the profile.
    gradnorm_path:
        The gradient norm along the profile.
    exitflag_path:
        The exitflags of the optimizer along the profile.
    time_path:
        The computation time of the optimizer runs along the profile.
    time_total:
        The total computation time for the profile.
    n_fval:
        Number of function evaluations.
    n_grad:
        Number of gradient evaluations.
    n_hess:
        Number of Hessian evaluations.
    message:
        Textual comment on the profile result.

    Notes
    -----
    Any field not supported by the profiler or the profiling optimizer is
    filled with None. Some fields are filled by pypesto itself.
    """

    def __init__(
        self,
        x_path: np.ndarray,
        fval_path: np.ndarray,
        ratio_path: np.ndarray,
        gradnorm_path: np.ndarray = np.nan,
        exitflag_path: np.ndarray = np.nan,
        time_path: np.ndarray = np.nan,
        time_total: float = 0.0,
        n_fval: int = 0,
        n_grad: int = 0,
        n_hess: int = 0,
        message: str = None,
    ):
        super().__init__()

        # initialize profile path
        x_shape = x_path.shape
        if len(x_shape) == 1:
            self.x_path = np.zeros((x_shape[0], 1))
            self.x_path[:, 0] = x_path[:]
        else:
            self.x_path = np.zeros((x_shape[0], x_shape[1]))
            self.x_path[:, :] = x_path[:, :]

        self.fval_path = np.asarray(fval_path)
        self.ratio_path = np.asarray(ratio_path)
        self.gradnorm_path = np.asarray(gradnorm_path)
        self.exitflag_path = np.asarray(exitflag_path)
        self.time_path = np.asarray(time_path)
        self.time_total = time_total
        self.n_fval = n_fval
        self.n_grad = n_grad
        self.n_hess = n_hess
        self.message = message

    def __getattr__(self, key):
        """Allow usage of keys like attributes."""
        try:
            return self[key]
        except KeyError:
            raise AttributeError(key)

    __setattr__ = dict.__setitem__
    __delattr__ = dict.__delitem__

    def append_profile_point(
        self,
        x: np.ndarray,
        fval: float,
        ratio: float,
        gradnorm: float = np.nan,
        time: float = np.nan,
        exitflag: float = np.nan,
        n_fval: int = 0,
        n_grad: int = 0,
        n_hess: int = 0,
    ) -> None:
        """
        Append a new point to the profile path.

        Parameters
        ----------
        x:
            The parameter values.
        fval:
            The function value at `x`.
        ratio:
            The ratio of the function value at `x` by the optimal function
            value.
        gradnorm:
            The gradient norm at `x`.
        time:
            The computation time to find `x`.
        exitflag:
            The exitflag of the optimizer (useful if an optimization was
            performed to find `x`).
        n_fval:
            Number of function evaluations performed to find `x`.
        n_grad:
            Number of gradient evaluations performed to find `x`.
        n_hess:
            Number of Hessian evaluations performed to find `x`.
        """
        # short function to append to numpy vectors
        def append_to_vector(field_name, val):
            field_new = np.zeros(self[field_name].size + 1)
            field_new[0:-1] = self[field_name]
            field_new[-1] = val
            self[field_name] = field_new

        # write profile path
        x_new = np.zeros((self.x_path.shape[0], self.x_path.shape[1] + 1))
        x_new[:, 0:-1] = self.x_path
        x_new[:, -1] = x
        self.x_path = x_new

        # append to other paths
        append_to_vector("fval_path", fval)
        append_to_vector("ratio_path", ratio)
        append_to_vector("gradnorm_path", gradnorm)
        append_to_vector("exitflag_path", exitflag)
        append_to_vector("time_path", time)

        # increment the time and f_eval counters
        self.time_total += time
        self.n_fval += n_fval
        self.n_grad += n_grad
        self.n_hess += n_hess

    def flip_profile(self) -> None:
        """
        Flip the profiling direction (left-right).

        Profiling direction needs to be changed once (if the profile is new),
        or twice if we append to an existing profile. All profiling paths
        are flipped in-place.
        """
        self.x_path = np.fliplr(self.x_path)
        self.fval_path = np.flip(self.fval_path)
        self.ratio_path = np.flip(self.ratio_path)
        self.gradnorm_path = np.flip(self.gradnorm_path)
        self.exitflag_path = np.flip(self.exitflag_path)
        self.time_path = np.flip(self.time_path)


class ProfileResult:
    """
    Result of the profile() function.

    It holds a list of profile lists. Each profile list consists of a list of
    `ProfilerResult` objects, one for each parameter.
    """

    def __init__(self):
        self.list = []

    def append_empty_profile_list(self) -> int:
        """Append an empty profile list to the list of profile lists.

        Returns
        -------
        index:
            The index of the created profile list.
        """
        self.list.append([])
        return len(self.list) - 1

    def append_profiler_result(
        self,
        profiler_result: ProfilerResult = None,
        profile_list: int = None,
    ) -> None:
        """Append the profiler result to the profile list.

        Parameters
        ----------
        profiler_result:
            The result of one profiler run for a parameter, or None if to be
            left empty.
        profile_list:
            Index specifying the profile list to which we want to append.
            Defaults to the last list.
        """
        if profile_list is None:
            profile_list = -1  # last
        profiler_result = copy.deepcopy(profiler_result)
        self.list[profile_list].append(profiler_result)

    def set_profiler_result(
        self,
        profiler_result: ProfilerResult,
        i_par: int,
        profile_list: int = None,
    ) -> None:
        """
        Write a profiler result to the result object.

        Parameters
        ----------
        profiler_result:
            The result of one (local) profiler run.
        i_par:
            Integer specifying the parameter index where to put
            profiler_result.
        profile_list:
            Index specifying the profile list. Defaults to the last list.
        """
        if profile_list is None:
            profile_list = -1  # last
        self.list[profile_list][i_par] = copy.deepcopy(profiler_result)

    def get_profiler_result(self, i_par: int, profile_list: int = None):
        """
        Get the profiler result at parameter index `i_par` of `profile_list`.

        Parameters
        ----------
        i_par:
            Integer specifying the profile index.
        profile_list:
            Index specifying the profile list. Defaults to the last list.
        """
        if profile_list is None:
            profile_list = -1  # last
        return self.list[profile_list][i_par]



import logging
from typing import Iterable, List, Optional, Tuple, Union

import matplotlib.pyplot as plt
import numpy as np
from matplotlib.ticker import MaxNLocator

from ..C import (
    RGBA,
    TRACE_X_STEPS,
    TRACE_X_TIME,
    TRACE_Y_FVAL,
    TRACE_Y_GRADNORM,
)
from ..history import HistoryBase
from ..result import Result
from .clust_color import assign_colors
from .misc import process_offset_y, process_result_list, process_y_limits
from .reference_points import ReferencePoint, create_references

logger = logging.getLogger(__name__)


def optimizer_history(
    results: Union[Result, List[Result]],
    ax: Optional[plt.Axes] = None,
    size: Tuple = (18.5, 10.5),
    trace_x: str = TRACE_X_STEPS,
    trace_y: str = TRACE_Y_FVAL,
    scale_y: str = 'log10',
    offset_y: Optional[float] = None,
    colors: Optional[Union[RGBA, List[RGBA]]] = None,
    y_limits: Optional[Union[float, List[float], np.ndarray]] = None,
    start_indices: Optional[Union[int, List[int]]] = None,
    reference: Optional[
        Union[ReferencePoint, dict, List[ReferencePoint], List[dict]]
    ] = None,
    legends: Optional[Union[str, List[str]]] = None,
) -> plt.Axes:
    """
    Plot history of optimizer.

    Can plot either the history of the cost function or of the gradient
    norm, over either the optimizer steps or the computation time.

    Parameters
    ----------
    results:
        Optimization result obtained by 'optimize.py' or list of those
    ax:
        Axes object to use.
    size: tuple, optional
        Figure size (width, height) in inches. Is only applied when no ax
        object is specified
    trace_x:
        What should be plotted on the x-axis?
        Possibilities: TRACE_X
        Default: TRACE_X_STEPS
    trace_y:
        What should be plotted on the y-axis?
        Possibilities: TRACE_Y_FVAL, TRACE_Y_GRADNORM
        Default: TRACE_Y_FVAl
    scale_y:
        May be logarithmic or linear ('log10' or 'lin')
    offset_y:
        Offset for the y-axis-values, as these are plotted on a log10-scale
        Will be computed automatically if necessary
    colors: list, or RGBA, optional
        list of colors, or single color
        color or list of colors for plotting. If not set, clustering is done
        and colors are assigned automatically
    y_limits:
        maximum value to be plotted on the y-axis, or y-limits
    start_indices:
        list of integers specifying the multistart to be plotted or
        int specifying up to which start index should be plotted
    reference:
        List of reference points for optimization results, containing at
        least a function value fval
    legends:
        Labels for line plots, one label per result object

    Returns
    -------
    ax:
        The plot axes.
    """
    if isinstance(start_indices, int):
        start_indices = list(range(start_indices))

    # parse input
    (results, colors, legends) = process_result_list(results, colors, legends)

    for j, result in enumerate(results):
        # extract cost function values from result
        vals = get_trace(result, trace_x, trace_y)

        # compute the necessary offset for the y-axis
        (vals, offset_y) = get_vals(
            vals, scale_y, offset_y, trace_y, start_indices
        )

        x_label, y_label = get_labels(trace_x, trace_y, offset_y)

        # call lowlevel plot routine
        ax = optimizer_history_lowlevel(
            vals,
            scale_y=scale_y,
            ax=ax,
            colors=colors[j],
            size=size,
            x_label=x_label,
            y_label=y_label,
            legend_text=legends[j],
        )

    # parse and apply plotting options
    ref = create_references(references=reference)

    # handle options
    ax = handle_options(ax, vals, trace_y, ref, y_limits, offset_y)

    return ax


def optimizer_history_lowlevel(
    vals: List[np.ndarray],
    scale_y: str = 'log10',
    colors: Optional[Union[RGBA, List[RGBA]]] = None,
    ax: Optional[plt.Axes] = None,
    size: Tuple = (18.5, 10.5),
    x_label: str = 'Optimizer steps',
    y_label: str = 'Objective value',
    legend_text: Optional[str] = None,
) -> plt.Axes:
    """
    Plot optimizer history using list of numpy arrays.

    Parameters
    ----------
    vals:
        list of 2xn-arrays (x_values and y_values of the trace)
    scale_y:
        May be logarithmic or linear ('log10' or 'lin')
    colors: list, or RGBA, optional
        list of colors, or single color
        color or list of colors for plotting. If not set, clustering is done
        and colors are assigned automatically
    ax:
        Axes object to use.
    size:
        see waterfall
    x_label:
        label for x-axis
    y_label:
        label for y-axis
    legend_text:
        Label for line plots

    Returns
    -------
    ax:
        The plot axes.
    """
    # axes
    if ax is None:
        ax = plt.subplots()[1]
        fig = plt.gcf()
        fig.set_size_inches(*size)

    # parse input
    fvals = []
    if isinstance(vals, list):
        # convert entries to numpy arrays
        for val in vals:
            # val is already an numpy array
            val = np.asarray(val)
            fvals.append(val[1, -1])
    else:
        # convert to a list of numpy arrays
        vals = np.asarray(vals)
        if vals.shape[0] != 2 or vals.ndim != 2:
            raise ValueError(
                'If numpy array is passed directly to lowlevel '
                'routine of optimizer_history, shape needs to '
                'be 2 x n.'
            )
        fvals = [vals[1, -1]]
        vals = [vals]
    n_fvals = len(fvals)

    # assign colors
    # note: this has to happen before sorting
    # to get the same colors in different plots
    colors = assign_colors(fvals, colors)

    # sort
    indices = sorted(range(n_fvals), key=lambda j: fvals[j])

    # plot
    ax.xaxis.set_major_locator(MaxNLocator(integer=True))
    for j, val in enumerate(vals):
        # collect and parse data
        j_fval = indices[j]
        color = colors[j_fval]
        if j == 0:
            tmp_legend = legend_text
        else:
            tmp_legend = None

        # line plots
        if scale_y == 'log10':
            ax.semilogy(val[0, :], val[1, :], color=color, label=tmp_legend)
        else:
            ax.plot(val[0, :], val[1, :], color=color, label=tmp_legend)

    # set labels
    ax.set_xlabel(x_label)
    ax.set_ylabel(y_label)
    ax.set_title('Optimizer history')
    if legend_text is not None:
        ax.legend()

    return ax


def get_trace(
    result: Result, trace_x: Optional[str], trace_y: Optional[str]
) -> List[np.ndarray]:
    """
    Get the values of the optimizer trace from the pypesto.Result object.

    Parameters
    ----------
    result: pypesto.Result
        Optimization result obtained by 'optimize.py'.
    trace_x: str, optional
        What should be plotted on the x-axis?
        Possibilities: TRACE_X
        Default: TRACE_X_STEPS
    trace_y: str, optional
        What should be plotted on the y-axis?
        Possibilities: TRACE_Y_FVAL, TRACE_Y_GRADNORM
        Default: TRACE_Y_FVAL

    Returns
    -------
    vals:
        list of (x,y)-values.
    x_label:
        label for x-axis to be plotted later.
    y_label:
        label for y-axis to be plotted later.
    """
    # get data frames
    histories: List[HistoryBase] = result.optimize_result.history

    vals = []

    for history in histories:
        options = history.options
        if trace_y == TRACE_Y_GRADNORM:
            # retrieve gradient trace, if saved
            if not options.trace_record or not options.trace_record_grad:
                raise ValueError("No gradient trace has been recorded.")
            grads = history.get_grad_trace()
            indices = [
                i
                for i, val in enumerate(grads)
                if val is not None and np.isfinite(val).all()
            ]

            grads = np.array([grads[i] for i in indices])

            # Get gradient trace, compute norm
            y_vals = np.linalg.norm(grads, axis=1)

        else:  # trace_y == TRACE_Y_FVAL:
            if not options.trace_record:
                raise ValueError("No function value trace has been recorded.")
            fvals = history.get_fval_trace()
            indices = [
                i
                for i, val in enumerate(fvals)
                if val is not None and np.isfinite(val)
            ]

            y_vals = np.array([fvals[i] for i in indices])

        # retrieve values from dataframe
        if trace_x == TRACE_X_TIME:
            times = np.array(history.get_time_trace())
            x_vals = times[indices]

        else:  # trace_x == TRACE_X_STEPS:
            x_vals = np.array(list(range(len(indices))))

        # write down values
        vals.append(np.vstack([x_vals, y_vals]))

    return vals


def get_vals(
    vals: List[np.ndarray],
    scale_y: Optional[str],
    offset_y: float,
    trace_y: str,
    start_indices: Iterable[int],
) -> Tuple[List[np.ndarray], float]:
    """
    Postprocess the values of the optimization history.

    Depending on the options set by the user (e.g. scale_y, offset_y,
    start_indices).

    Parameters
    ----------
    vals:
        list of numpy arrays of dimension 2 x len(start_indices)
    scale_y:
        May be logarithmic or linear ('log10' or 'lin')
    offset_y:
        offset for the y-axis, as this is supposed to be in log10-scale
    trace_y:
        What should be plotted on the y-axis
    start_indices:
        list of integers specifying the multi start indices to be plotted

    Returns
    -------
    vals:
        list of numpy arrays
    offset_y:
        offset for the y-axis, if this is supposed to be in log10-scale
    y_label:
        Label for y axis
    """
    # get list of indices
    if start_indices is None:
        start_indices = np.array(range(len(vals)))
    else:
        # check whether list or maximum value
        start_indices = np.array(start_indices)

        # check, whether index set is not too big
        existing_indices = np.array(range(len(vals)))
        start_indices = np.intersect1d(start_indices, existing_indices)

    # reduce values to listed values
    vals = [val for i, val in enumerate(vals) if i in start_indices]

    # get the minimal value which should be plotted
    min_val = np.inf
    for val in vals:
        tmp_min = np.min(val[1, :])
        min_val = np.min([min_val, tmp_min])

    # check, whether offset can be used with this data
    if trace_y == TRACE_Y_FVAL:
        offset_y = process_offset_y(offset_y, scale_y, min_val)
    else:
        offset_y = 0.0

    if offset_y != 0:
        for val in vals:
            val[1, :] += offset_y * np.ones(val[1].shape)

    return vals, offset_y


def get_labels(trace_x: str, trace_y: str, offset_y: float) -> Tuple[str, str]:
    """
    Generate labels for x and y axes of the history plot.

    Parameters
    ----------
    trace_x:
        What should be plotted on the x-axis. Possible values: TRACE_X.
    trace_y:
        What should be plotted on the y-axis.
        Possible values: TRACE_Y_FVAL, TRACE_Y_GRADNORM.
    offset_y:
        Offset for the y-axis-values.
    Returns
    -------
    labels for x and y axes

    """
    x_label = ''
    y_label = ''

    if trace_x == TRACE_X_TIME:
        x_label = 'Computation time [s]'
    else:
        x_label = 'Optimizer steps'

    if trace_y == TRACE_Y_GRADNORM:
        y_label = 'Gradient norm'
    else:
        y_label = 'Objective value'

    if offset_y != 0:
        y_label = 'Offsetted ' + y_label.lower()

    return x_label, y_label


def handle_options(
    ax: plt.Axes,
    vals: List[np.ndarray],
    trace_y: str,
    ref: List[ReferencePoint],
    y_limits: Union[float, np.ndarray, None],
    offset_y: float,
) -> plt.Axes:
    """
    Apply post-plotting transformations to the axis object.

    Get the limits for the y-axis, plots the reference points, will do
    more at a later time point.

    Parameters
    ----------
    ref:
        List of reference points for optimization results, containing et
        least a function value fval
    vals:
        list of numpy arrays of size 2 x number of values
    trace_y:
        What should be plotted on the x-axis.
    ax:
        Axes object to use.
    y_limits:
        maximum value to be plotted on the y-axis, or y-limits
    offset_y:
        offset for the y-axis, if this is supposed to be in log10-scale

    Returns
    -------
    ax:
        The plot axes.
    """
    # handle y-limits
    ax = process_y_limits(ax, y_limits)

    if trace_y == TRACE_Y_FVAL:
        # handle reference points
        if len(ref) > 0:
            # plot reference points
            # get length of longest trajectory
            max_len = 0
            for val in vals:
                max_len = np.max([max_len, val[0, -1]])

            for i_ref in ref:
                ax.plot(
                    [0, max_len],
                    [i_ref.fval + offset_y, i_ref.fval + offset_y],
                    '--',
                    color=i_ref.color,
                    label=i_ref.legend,
                )

                # create legend for reference points
                if i_ref.legend is not None:
                    ax.legend()
    else:
        logger.warning(
            f'Reference point is currently only implemented for trace_y == '
            f'{TRACE_Y_FVAL} and will not be plotted for trace_y == {trace_y}.'
        )

    return ax



from typing import Sequence, Union

import matplotlib.axes
import matplotlib.pyplot as plt
import numpy as np

from ..profile import calculate_approximate_ci, chi2_quantile_to_ratio
from ..result import Result


def profile_cis(
    result: Result,
    confidence_level: float = 0.95,
    profile_indices: Sequence[int] = None,
    profile_list: int = 0,
    color: Union[str, tuple] = 'C0',
    show_bounds: bool = False,
    ax: matplotlib.axes.Axes = None,
) -> matplotlib.axes.Axes:
    """
    Plot approximate confidence intervals based on profiles.

    Parameters
    ----------
    result:
        The result object after profiling.
    confidence_level:
        The confidence level in (0,1), which is translated to an approximate
        threshold assuming a chi2 distribution, using
        `pypesto.profile.chi2_quantile_to_ratio`.
    profile_indices:
        List of integer values specifying which profiles should be plotted.
        Defaults to the indices for which profiles were generated in profile
        list `profile_list`.
    profile_list:
        Index of the profile list to be used.
    color:
        Main plot color.
    show_bounds:
        Whether to show, and extend the plot to, the lower and upper bounds.
    ax:
        Axes object to use. Default: Create a new one.
    """
    # extract problem
    problem = result.problem
    # extract profile list
    profile_list = result.profile_result.list[profile_list]

    if profile_indices is None:
        profile_indices = [ix for ix, res in enumerate(profile_list) if res]

    if ax is None:
        _, ax = plt.subplots()

    confidence_ratio = chi2_quantile_to_ratio(confidence_level)

    # calculate intervals
    intervals = []
    for i_par in range(problem.dim_full):
        if i_par not in profile_indices:
            continue
        xs = profile_list[i_par].x_path[i_par]
        ratios = profile_list[i_par].ratio_path
        lb, ub = calculate_approximate_ci(
            xs=xs, ratios=ratios, confidence_ratio=confidence_ratio
        )
        intervals.append((lb, ub))

    x_names = [problem.x_names[ix] for ix in profile_indices]

    for ix, (lb, ub) in enumerate(intervals):
        ax.plot([lb, ub], [ix + 1, ix + 1], marker='|', color=color)

    parameters_ind = np.arange(1, len(intervals) + 1)
    ax.set_yticks(parameters_ind)
    ax.set_yticklabels(x_names)
    ax.set_ylabel("Parameter")
    ax.set_xlabel("Parameter value")

    if show_bounds:
        lb = problem.lb_full[profile_indices]
        ax.plot(lb, parameters_ind, 'k--', marker='+')
        ub = problem.ub_full[profile_indices]
        ax.plot(ub, parameters_ind, 'k--', marker='+')

    return ax



"""Convenience functions for working with HDF5 files."""

from numbers import Integral, Number, Real
from typing import Collection

import h5py
import numpy as np


def write_array(f: h5py.Group, path: str, values: Collection) -> None:
    """
    Write array to hdf5.

    Parameters
    ----------
    f:
        h5py.Group  where dataset should be created
    path:
        path of the dataset to create
    values:
        array to write
    """
    if all(isinstance(x, Integral) for x in values):
        write_int_array(f, path, values)
    elif all(isinstance(x, Real) for x in values):
        write_float_array(f, path, values)
    elif all(isinstance(x, str) for x in values):
        write_string_array(f, path, values)
    else:
        f[path] = values


def write_string_array(f: h5py.Group, path: str, strings: Collection) -> None:
    """
    Write string array to hdf5.

    Parameters
    ----------
    f:
        h5py.Group where dataset should be created
    path:
        path of the dataset to create
    strings:
        list of strings to be written to f
    """
    dt = h5py.special_dtype(vlen=str)
    dset = f.create_dataset(path, (len(strings),), dtype=dt)
    dset[:] = [s.encode('utf8') for s in strings]


def write_float_array(
    f: h5py.Group, path: str, values: Collection[Number], dtype='f8'
) -> None:
    """
    Write float array to hdf5.

    Parameters
    ----------
    f:
        h5py.Group where dataset should be created
    path:
        path of the dataset to create
    values:
        array to write
    dtype:
        datatype
    """
    if path not in f:
        dset = f.create_dataset(path, (np.shape(values)), dtype=dtype)
    else:
        dset = f[path]
    dset[:] = values


def write_int_array(
    f: h5py.Group, path: str, values: Collection[int], dtype='<i4'
):
    """
    Write integer array to hdf5.

    Parameters
    ----------
    f:
        h5py.Group where dataset should be created
    path:
        path of the dataset to create
    values:
        array to write
    dtype:
        datatype
    """
    dset = f.create_dataset(path, (len(values),), dtype=dtype)
    dset[:] = values



"""
History
=======

Objetive function call history. The history tracks and stores function
evaluations performed by e.g. the optimizer and other routines, allowing to
e.g. recover results from failed runs, fill in further details,
and evaluate performance.
"""

from .base import CountHistory, CountHistoryBase, HistoryBase, NoHistory
from .csv import CsvHistory
from .generate import create_history
from .hdf5 import Hdf5History
from .memory import MemoryHistory
from .optimizer import OptimizerHistory
from .options import HistoryOptions
from .util import CsvHistoryTemplateError, HistoryTypeError



"""Engines with multi-node parallelization."""
import logging
from typing import List

import cloudpickle as pickle
from mpi4py import MPI
from mpi4py.futures import MPIPoolExecutor
from tqdm import tqdm

from .base import Engine
from .task import Task

logger = logging.getLogger(__name__)


def work(pickled_task):
    """Unpickle and execute task."""
    task = pickle.loads(pickled_task)
    return task.execute()


class MPIPoolEngine(Engine):
    """
    Parallelize the task execution.

    Uses `mpi4py <https://mpi4py.readthedocs.io/en/stable/>`_.
    To be called with:
    ``mpiexec -np #Workers+1 python -m mpi4py.futures YOURFILE.py``
    """

    def __init__(self):
        super().__init__()

    def execute(self, tasks: List[Task], progress_bar: bool = True):
        """
        Pickle tasks and distribute work to workers.

        Parameters
        ----------
        tasks:
            List of tasks to execute.
        progress_bar:
            Whether to display a progress bar.
        """
        pickled_tasks = [pickle.dumps(task) for task in tasks]

        n_procs = MPI.COMM_WORLD.Get_size()  # Size of communicator
        logger.info(
            f"Performing parallel task execution on {n_procs-1} "
            f"workers with one manager."
        )

        with MPIPoolExecutor() as executor:
            results = executor.map(
                work, tqdm(pickled_tasks, disable=not progress_bar)
            )
        return results



import copy
from typing import Dict, List, Sequence, Union

import numpy as np
from tqdm import tqdm

from ..problem import Problem
from ..result import McmcPtResult
from .sampler import InternalSampler, Sampler


class ParallelTemperingSampler(Sampler):
    """Simple parallel tempering sampler."""

    def __init__(
        self,
        internal_sampler: InternalSampler,
        betas: Sequence[float] = None,
        n_chains: int = None,
        options: Dict = None,
    ):
        super().__init__(options)

        # set betas
        if (betas is None) == (n_chains is None):
            raise ValueError("Set either betas or n_chains.")
        if betas is None:
            betas = near_exponential_decay_betas(
                n_chains=n_chains,
                exponent=self.options['exponent'],
                max_temp=self.options['max_temp'],
            )
        if betas[0] != 1.0:
            raise ValueError("The first chain must have beta=1.0")
        self.betas0 = np.array(betas)
        self.betas = None

        self.temper_lpost = self.options['temper_log_posterior']

        self.samplers = [
            copy.deepcopy(internal_sampler) for _ in range(len(self.betas0))
        ]
        # configure internal samplers
        for sampler in self.samplers:
            sampler.make_internal(temper_lpost=self.temper_lpost)

    @classmethod
    def default_options(cls) -> Dict:
        """Return the default options for the sampler."""
        return {
            'max_temp': 5e4,
            'exponent': 4,
            'temper_log_posterior': False,
            'show_progress': True,
        }

    def initialize(
        self, problem: Problem, x0: Union[np.ndarray, List[np.ndarray]]
    ):
        """Initialize all samplers."""
        n_chains = len(self.samplers)
        if isinstance(x0, list):
            x0s = x0
        else:
            x0s = [x0 for _ in range(n_chains)]
        for sampler, x0 in zip(self.samplers, x0s):
            _problem = copy.deepcopy(problem)
            sampler.initialize(_problem, x0)
        self.betas = self.betas0

    def sample(self, n_samples: int, beta: float = 1.0):
        """Sample and swap in between samplers."""
        show_progress = self.options['show_progress']
        # loop over iterations
        for i_sample in tqdm(range(int(n_samples)), disable=not show_progress):
            # TODO test
            # sample
            for sampler, beta in zip(self.samplers, self.betas):
                sampler.sample(n_samples=1, beta=beta)

            # swap samples
            swapped = self.swap_samples()

            # adjust temperatures
            self.adjust_betas(i_sample, swapped)

    def get_samples(self) -> McmcPtResult:
        """Concatenate all chains."""
        results = [sampler.get_samples() for sampler in self.samplers]
        trace_x = np.array([result.trace_x[0] for result in results])
        trace_neglogpost = np.array(
            [result.trace_neglogpost[0] for result in results]
        )
        trace_neglogprior = np.array(
            [result.trace_neglogprior[0] for result in results]
        )
        return McmcPtResult(
            trace_x=trace_x,
            trace_neglogpost=trace_neglogpost,
            trace_neglogprior=trace_neglogprior,
            betas=self.betas,
        )

    def swap_samples(self) -> Sequence[bool]:
        """Swap samples as in Vousden2016."""
        # for recording swaps
        swapped = []

        if len(self.betas) == 1:
            # nothing to be done
            return swapped

        # beta differences
        dbetas = self.betas[:-1] - self.betas[1:]

        # loop over chains from highest temperature down
        for dbeta, sampler1, sampler2 in reversed(
            list(zip(dbetas, self.samplers[:-1], self.samplers[1:]))
        ):
            # extract samples
            sample1 = sampler1.get_last_sample()
            sample2 = sampler2.get_last_sample()

            # extract log likelihood values
            sample1_llh = sample1.lpost - sample1.lprior
            sample2_llh = sample2.lpost - sample2.lprior

            # swapping probability
            p_acc_swap = dbeta * (sample2_llh - sample1_llh)

            # flip a coin
            u = np.random.uniform(0, 1)

            # check acceptance
            swap = np.log(u) < p_acc_swap
            if swap:
                # swap
                sampler2.set_last_sample(sample1)
                sampler1.set_last_sample(sample2)

            # record
            swapped.insert(0, swap)
        return swapped

    def adjust_betas(self, i_sample: int, swapped: Sequence[bool]):
        """Adjust temperature values. Default: Do nothing."""


def near_exponential_decay_betas(
    n_chains: int, exponent: float, max_temp: float
) -> np.ndarray:
    """Initialize betas in a near-exponential decay scheme.

    Parameters
    ----------
    n_chains:
        Number of chains to use.
    exponent:
        Decay exponent. The higher, the more small temperatures are used.
    max_temp:
        Maximum chain temperature.
    """
    # special case of one chain
    if n_chains == 1:
        return np.array([1.0])

    temperatures = (
        np.linspace(1, max_temp ** (1 / exponent), n_chains) ** exponent
    )
    betas = 1 / temperatures

    return betas



from copy import deepcopy
from typing import Sequence, Tuple

import numpy as np

from ..C import FVAL, GRAD, HESS, HESSP, RDATAS, RES, SRES, ModeType
from .base import ObjectiveBase, ResultDict


class AggregatedObjective(ObjectiveBase):
    """Aggregates multiple objectives into one objective."""

    def __init__(
        self,
        objectives: Sequence[ObjectiveBase],
        x_names: Sequence[str] = None,
    ):
        """
        Initialize objective.

        Parameters
        ----------
        objectives:
            Sequence of pypesto.ObjectiveBase instances
        x_names:
            Sequence of names of the (optimized) parameters.
            (Details see documentation of x_names in
            :class:`pypesto.ObjectiveBase`)
        """
        # input typechecks
        if not isinstance(objectives, Sequence):
            raise TypeError(
                f'Objectives must be a Sequence, ' f'was {type(objectives)}.'
            )

        if not all(
            isinstance(objective, ObjectiveBase) for objective in objectives
        ):
            raise TypeError(
                'Objectives must only contain elements of type'
                'pypesto.Objective'
            )

        if not objectives:
            raise ValueError('Length of objectives must be at least one')

        self._objectives = objectives

        super().__init__(x_names=x_names)

    def __deepcopy__(self, memodict=None):
        """Create copy of objective."""
        other = AggregatedObjective(
            objectives=[deepcopy(objective) for objective in self._objectives],
            x_names=deepcopy(self.x_names),
        )
        for key in set(self.__dict__.keys()) - {'_objectives', 'x_names'}:
            other.__dict__[key] = deepcopy(self.__dict__[key])

        return other

    def check_mode(self, mode: ModeType) -> bool:
        """See `ObjectiveBase` documentation."""
        return all(
            objective.check_mode(mode) for objective in self._objectives
        )

    def check_sensi_orders(
        self,
        sensi_orders: Tuple[int, ...],
        mode: ModeType,
    ) -> bool:
        """See `ObjectiveBase` documentation."""
        return all(
            objective.check_sensi_orders(sensi_orders, mode)
            for objective in self._objectives
        )

    def call_unprocessed(
        self,
        x: np.ndarray,
        sensi_orders: Tuple[int, ...],
        mode: ModeType,
        **kwargs,
    ) -> ResultDict:
        """
        See `ObjectiveBase` for more documentation.

        Main method to overwrite from the base class. It handles and
        delegates the actual objective evaluation.
        """
        return aggregate_results(
            [
                objective.call_unprocessed(x, sensi_orders, mode, **kwargs)
                for objective in self._objectives
            ]
        )

    def initialize(self):
        """See `ObjectiveBase` documentation."""
        for objective in self._objectives:
            objective.initialize()

    def get_config(self) -> dict:
        """Return basic information of the objective configuration."""
        info = super().get_config()
        for n_obj, obj in enumerate(self._objectives):
            info[f'objective_{n_obj}'] = obj.get_config()
        return info


def aggregate_results(rvals: Sequence[ResultDict]) -> ResultDict:
    """
    Aggregate the results from the provided ResultDicts into a single one.

    Parameters
    ----------
    rvals:
        results to aggregate
    """
    # sum over fval/grad/hess, if available in all rvals
    result = {
        key: sum(rval[key] for rval in rvals)
        for key in [FVAL, GRAD, HESS, HESSP]
        if all(key in rval for rval in rvals)
    }

    # extract rdatas and flatten
    result[RDATAS] = []
    for rval in rvals:
        if RDATAS in rval:
            result[RDATAS].extend(rval[RDATAS])

    # initialize res and sres
    if RES in rvals[0]:
        res = np.asarray(rvals[0][RES])
    else:
        res = None

    if SRES in rvals[0]:
        sres = np.asarray(rvals[0][SRES])
    else:
        sres = None

    # skip iobj=0 after initialization, stack matrices
    for rval in rvals[1:]:
        if res is not None:
            res = np.hstack([res, np.asarray(rval[RES])])
        if sres is not None:
            sres = np.vstack([sres, np.asarray(rval[SRES])])

    # fill res, sres into result
    if res is not None:
        result[RES] = res
    if sres is not None:
        result[SRES] = sres

    return result



"""Interface to Julia via pyjulia."""

from typing import Callable, Union

import numpy as np

from ..function import Objective


def _as_array(input):
    """Convert output to numpy array."""
    if callable(input):

        def wrapper(*args, **kwargs):
            return np.asarray(input(*args, **kwargs))

        return wrapper

    return np.asarray(input)


def _read_source(module_name: str, source_file: str) -> None:
    """Read source if module not attached to julia Main yet.

    Parameters
    ----------
    module_name: Julia module name.
    source_file: Qualified Julia source file.
    """
    from julia import Main

    if not hasattr(Main, module_name):
        Main.include(source_file)


class JuliaObjective(Objective):
    """Wrapper around an objective defined in Julia.

    This class provides objective function wrappers around Julia objects.
    It expects the corresponding Julia objects to be defined in a
    `source_file` within a `module`.

    We use the PyJulia package to access Julia from inside Python.
    It can be installed via `pip install pypesto[julia]`, however requires
    additional Julia dependencies to be installed via:

    >>> python -c "import julia; julia.install()"

    For further information, see
    https://pyjulia.readthedocs.io/en/latest/installation.html.

    There are some known problems, e.g. with statically linked Python
    interpreters, see
    https://pyjulia.readthedocs.io/en/latest/troubleshooting.html
    for details.
    Possible solutions are to pass ``compiled_modules=False`` to the Julia
    constructor early in your code:

    >>> from julia.api import Julia
    >>> jl = Julia(compiled_modules=False)

    This however slows down loading and using Julia packages, especially for
    large ones.
    An alternative is to use the ``python-jl`` command shipped with PyJulia:

    >>> python-jl MY_SCRIPT.py

    This basically launches a Python interpreter inside Julia.
    When using Jupyter notebooks, this wrapper can be installed as an
    additional kernel via:

    >>> python -m ipykernel install --name python-jl [--prefix=/path/to/python/env]

    And changing the first argument in
    ``/path/to/python/env/share/jupyter/kernels/python-jl/kernel.json``
    to ``python-jl``.

    Model simulations are eagerly converted to Python objects
    (specifically, `numpy.ndarray` and `pandas.DataFrame`).
    This can introduce overhead and could be avoided by an alternative
    lazy implementation.

    Parameters
    ----------
    module:
        Julia module name.
    source_file:
        Julia source file name. Defaults to `{module}.jl`.
    fun, grad, hess, res, sres:
        Names of callables within the Julia code of the corresponding
        objective functions and derivatives.
    """

    def __init__(
        self,
        module: str,
        source_file: str = None,
        fun: str = None,
        grad: str = None,
        hess: str = None,
        res: str = None,
        sres: str = None,
    ):
        # lazy imports
        try:
            from julia import Main  # noqa: F401
        except ImportError:
            raise ImportError(
                "Install PyJulia, e.g. via `pip install pypesto[julia]`, "
                "and see the class documentation",
            )

        # store module name and source file
        self.module: str = module
        if source_file is None:
            source_file = module + ".jl"
        self.source_file: str = source_file
        _read_source(self.module, self.source_file)

        # store function names
        self._fun: str = fun
        self._grad: str = grad
        self._hess: str = hess
        self._res: str = res
        self._sres: str = sres

        # get callables
        fun, grad, hess, res, sres = self._get_callables()
        super().__init__(fun=fun, grad=grad, hess=hess, res=res, sres=sres)

    def get(self, name: str, as_array: bool = False) -> Union[Callable, None]:
        """Get variable from Julia module.

        Use this function to access any variable from the Julia module.
        """
        from julia import Main

        if name is not None:
            ret = getattr(getattr(Main, self.module), name, None)
            if as_array:
                ret = _as_array(ret)
            return ret
        return None

    def _get_callables(self) -> tuple:
        """Get all callables."""
        fun = self.get(self._fun)
        grad = self.get(self._grad, as_array=True)
        hess = self.get(self._hess, as_array=True)
        res = self.get(self._res, as_array=True)
        sres = self.get(self._sres, as_array=True)
        return fun, grad, hess, res, sres

    def __getstate__(self):
        return {
            "module": self.module,
            "source_file": self.source_file,
            "_fun": self._fun,
            "_grad": self._grad,
            "_hess": self._hess,
            "_res": self._res,
            "_sres": self._sres,
        }

    def __setstate__(self, d):
        for key, val in d.items():
            setattr(self, key, val)
        _read_source(self.module, self.source_file)

        fun, grad, hess, res, sres = self._get_callables()
        super().__init__(fun=fun, grad=grad, hess=hess, res=res, sres=sres)

    def __deepcopy__(self, memodict=None) -> 'JuliaObjective':
        return JuliaObjective(
            module=self.module,
            source_file=self.source_file,
            fun=self._fun,
            grad=self._grad,
            hess=self._hess,
            res=self._res,
            sres=self._sres,
        )


def display_source_ipython(source_file: str):
    """Display source code as syntax highlighted HTML within IPython."""
    import IPython.display as display
    from pygments import highlight
    from pygments.formatters import HtmlFormatter
    from pygments.lexers import JuliaLexer

    with open(source_file) as f:
        code = f.read()

    formatter = HtmlFormatter()
    return display.HTML(
        '<style type="text/css">{}</style>{}'.format(
            formatter.get_style_defs('.highlight'),
            highlight(code, JuliaLexer(), formatter),
        )
    )



"""
PEtab
=====

pyPESTO support for the PEtab data format.
"""
import warnings

from .importer import PetabImporter
from .pysb_importer import PetabImporterPysb

# PEtab and amici are optional dependencies


try:
    import petab
except ImportError:
    warnings.warn(
        "PEtab import requires an installation of petab "
        "(https://github.com/PEtab-dev/PEtab). "
        "Install via `pip3 install petab`."
    )
try:
    import amici
except ImportError:
    warnings.warn(
        "PEtab import requires an installation of amici "
        "(https://github.com/AMICI-dev/AMICI). "
        "Install via `pip3 install amici`."
    )



"""Startpoint generation utilities."""

import numpy as np


def rescale(points, lb: np.ndarray, ub: np.ndarray) -> np.ndarray:
    """Rescale points from [0, 1] to [lb, ub].

    Parameters
    ----------
    points: ndarray, shape=(n_starts, dim)
        Points in bounds [lb, ub]
    lb, ub: ndarray, shape=(1, dim)
        The boundaries, all components must be finite.
    """
    rescaled_points = points * (ub - lb) + lb
    return rescaled_points



"""Calibrate a PEtab Select model with pyPESTO."""
import time
from typing import Any, Callable, Dict, List, Optional

from petab_select import Criterion, Model

from ..C import TYPE_POSTPROCESSOR
from ..objective import ObjectiveBase
from ..optimize import minimize
from ..problem import Problem
from ..result import OptimizerResult, Result
from .misc import model_to_pypesto_problem

OBJECTIVE_CUSTOMIZER_TYPE = Callable[[ObjectiveBase], None]
POSTPROCESSOR_TYPE = Callable[["ModelProblem"], None]


class ModelProblem:
    """Handles all required calibration tasks on a model.

    Handles the creation, estimation, and evaluation of a model. Here, a model
    is a PEtab problem that is patched with a dictionary of custom parameter
    values (which may specify that the parameter should be estimated).
    Evaluation refers to criterion values such as AIC.

    Attributes
    ----------
    best_start:
        The best start from a pyPESTO optimize result.
    criterion:
        The criterion that should be computed after the model is
        calibrated.
    minimize_options:
        Keyword argument options that will be passed on to
        `pypesto.optimize.minimize`.
    minimize_result:
        A pyPESTO result with an optimize result.
    model:
        A PEtab Select model.
    model_id:
        The ID of the PEtab Select model.
    objective_customizer:
        A method that takes a `pypesto.objective.AmiciObjective` as
        input, and makes changes to the objective in-place.
    postprocessor:
        A method that takes a `ModelSelectionProblem` as input. For
        example, this can be a function that generates a waterfall
        plot. This postprocessor is applied at the end of the
        `ModelProblem.set_result` method.
    pypesto_problem:
        The pyPESTO problem for the model.
    valid:
        If `False`, the model will not be tested.
    x_guess:
        A single startpoint, that will be used as one of the
        startpoints in the multi-start optimization.
    """

    def __init__(
        self,
        model: Model,
        criterion: Criterion,
        valid: bool = True,
        autorun: bool = True,
        x_guess: List[float] = None,
        minimize_options: Dict = None,
        objective_customizer: Optional[OBJECTIVE_CUSTOMIZER_TYPE] = None,
        postprocessor: Optional[TYPE_POSTPROCESSOR] = None,
        model_to_pypesto_problem_method: Callable[[Any], Problem] = None,
    ):
        """Construct then calibrate a model problem.

        See the class documentation for documentation of most parameters.

        Parameters
        ----------
        autorun:
            If `False`, the model parameters will not be estimated. Allows
            users to manually call pypesto.minimize with custom options,
            then`set_result()`.

        TODO: constraints
        """
        self.model = model
        self.criterion = criterion
        self.valid = valid

        self.minimize_options = {}
        if minimize_options is not None:
            self.minimize_options = minimize_options

        self.model_id = self.model.model_id
        self.objective_customizer = objective_customizer
        self.postprocessor = postprocessor
        self.best_start = None
        self.minimize_result = None
        self.x_guess = x_guess

        self.model_to_pypesto_problem_method = model_to_pypesto_problem_method
        if model_to_pypesto_problem_method is None:
            self.model_to_pypesto_problem_method = model_to_pypesto_problem

        if self.valid:
            self.pypesto_problem = self.model_to_pypesto_problem_method(
                self.model,
                x_guesses=None if self.x_guess is None else [self.x_guess],
            )

            if self.objective_customizer is not None:
                self.objective_customizer(self.pypesto_problem.objective)

            # TODO autorun may be unnecessary now that the `minimize_options`
            # argument is implemented.
            if autorun:
                # If there are no estimated parameters, evaluate the objective
                # function and generate a fake optimization result.
                if not self.pypesto_problem.x_free_indices:
                    fake_result_start_time = time.time()
                    fake_result_fval = self.pypesto_problem.objective([])
                    fake_result_evaluation_time = (
                        time.time() - fake_result_start_time
                    )
                    self.set_result(
                        create_fake_pypesto_result_from_fval(
                            fval=fake_result_fval,
                            evaluation_time=fake_result_evaluation_time,
                        )
                    )
                # TODO rename `minimize_options` to `minimize_kwargs`.
                # TODO or allow users to provide custom `minimize` methods?
                else:
                    self.set_result(
                        minimize(
                            self.pypesto_problem,
                            **minimize_options,
                        )
                    )

    def set_result(self, result: Result):
        """Postprocess a result.

        Parameters
        ----------
        result:
            A pyPESTO result with an optimize result.
        """
        self.minimize_result = result
        # TODO extract best parameter estimates, to use as start point for
        # subsequent models in model selection, for parameters in those models
        # that were estimated in this model.
        self.best_start = self.minimize_result.optimize_result.list[0]
        self.model.set_criterion(Criterion.NLLH, float(self.best_start.fval))
        self.model.compute_criterion(criterion=self.criterion)

        self.model.estimated_parameters = {
            id: float(value)
            for index, (id, value) in enumerate(
                dict(
                    zip(
                        self.pypesto_problem.x_names,
                        self.best_start.x,
                    )
                ).items()
            )
            if index in self.pypesto_problem.x_free_indices
        }

        if self.postprocessor is not None:
            self.postprocessor(self)


def create_fake_pypesto_result_from_fval(
    fval: float,
    evaluation_time: float = 0.0,
) -> Result:
    """Create a result for problems with no estimated parameters.

    Parameters
    ----------
    fval:
        The objective function value.
    """
    result = Result()

    optimizer_result = OptimizerResult(
        id="fake_result_for_problem_with_no_estimated_parameters",
        x=[],
        fval=fval,
        grad=None,
        hess=None,
        res=None,
        sres=None,
        n_fval=1,
        n_grad=0,
        n_hess=0,
        n_res=0,
        n_sres=0,
        x0=[],
        fval0=fval,
        history=None,
        exitflag=0,
        time=evaluation_time,
        message="Fake result for problem with no estimated parameters.",
    )

    result.optimize_result.append(optimizer_result)
    return result



"""Latin hypercube sampling."""

import numpy as np

from .base import CheckedStartpoints
from .util import rescale


def latin_hypercube(
    n_starts: int,
    lb: np.ndarray,
    ub: np.ndarray,
    smooth: bool = True,
) -> np.ndarray:
    """Generate latin hypercube points.

    Parameters
    ----------
    n_starts:
        Number of points.
    lb:
        Lower bound.
    ub:
        Upper bound.
    smooth:
        Whether a (uniformly chosen) random starting point within the
        hypercube [i/n_starts, (i+1)/n_starts] should be chosen (True) or
        the midpoint of the interval (False).

    Returns
    -------
    xs:
        Latin hypercube points, shape (n_starts, n_x).
    """
    if not np.isfinite(ub).all() or not np.isfinite(lb).all():
        raise ValueError(
            "Cannot use latin hypercube startpoint method with non-finite "
            "boundaries.",
        )

    # parse
    dim = lb.size
    lb = lb.reshape((1, -1))
    ub = ub.reshape((1, -1))

    # sample
    xs = _latin_hypercube_unit(n_starts, dim, smooth)

    # re-scale
    xs = rescale(xs, lb, ub)

    return xs


def _latin_hypercube_unit(
    n_starts: int,
    dim: int,
    smooth: bool,
) -> np.ndarray:
    """Generate simple latin hypercube points in [0, 1].

    Parameters are as for `latin_hypercube`.

    Returns
    -------
    xs: Latin hypercube points sampled in [0, 1], shape (n_starts, dim).
    """
    xs = np.empty((n_starts, dim))

    for i_dim in range(dim):
        xs[:, i_dim] = np.random.permutation(np.arange(n_starts))

    if smooth:
        xs += np.random.random((n_starts, dim))
    else:
        xs += 0.5

    xs /= n_starts

    return xs


class LatinHypercubeStartpoints(CheckedStartpoints):
    """Generate latin hypercube-sampled startpoints.

    See e.g. https://en.wikipedia.org/wiki/Latin_hypercube_sampling.
    """

    def __init__(
        self,
        use_guesses: bool = True,
        check_fval: bool = False,
        check_grad: bool = False,
        smooth: bool = True,
    ):
        """Initialize.

        Parameters
        ----------
        use_guesses, check_fval, check_grad:
            As in CheckedStartpoints.
        smooth:
            Whether a (uniformly chosen) random starting point within the
            hypercube [i/n_starts, (i+1)/n_starts] should be chosen (True) or
            the midpoint of the interval (False).
        """
        super().__init__(
            use_guesses=use_guesses,
            check_fval=check_fval,
            check_grad=check_grad,
        )
        self.smooth: bool = smooth

    def sample(
        self,
        n_starts: int,
        lb: np.ndarray,
        ub: np.ndarray,
    ) -> np.ndarray:
        """Call function."""
        return latin_hypercube(
            n_starts=n_starts,
            lb=lb,
            ub=ub,
            smooth=self.smooth,
        )



from typing import Tuple, Union

import numpy as np

from .ensemble import Ensemble, EnsemblePrediction
from .util import get_prediction_dataset


def get_covariance_matrix_parameters(ens: Ensemble) -> np.ndarray:
    """
    Compute the covariance of ensemble parameters.

    Parameters
    ----------
    ens:
        Ensemble object containing a set of parameter vectors

    Returns
    -------
    covariance_matrix:
        covariance matrix of ensemble parameters
    """
    # call lowlevel routine using the parameter ensemble
    return np.cov(ens.x_vectors.transpose())


def get_covariance_matrix_predictions(
    ens: Union[Ensemble, EnsemblePrediction], prediction_index: int = 0
) -> np.ndarray:
    """
    Compute the covariance of ensemble predictions.

    Parameters
    ----------
    ens:
        Ensemble object containing a set of parameter vectors and a set of
        predictions or EnsemblePrediction object containing only predictions
    prediction_index:
        index telling which prediction from the list should be analyzed

    Returns
    -------
    covariance_matrix:
        covariance matrix of ensemble predictions
    """
    # extract the an array of predictions from either an Ensemble object or an
    # EnsemblePrediction object
    dataset = get_prediction_dataset(ens, prediction_index)

    # call lowlevel routine using the prediction ensemble
    return np.cov(dataset)


def get_spectral_decomposition_parameters(
    ens: Ensemble,
    normalize: bool = False,
    only_separable_directions: bool = False,
    cutoff_absolute_separable: float = 1e-16,
    cutoff_relative_separable: float = 1e-16,
    only_identifiable_directions: bool = False,
    cutoff_absolute_identifiable: float = 1e-16,
    cutoff_relative_identifiable: float = 1e-16,
) -> Tuple[np.ndarray, np.ndarray]:
    """
    Compute the spectral decomposition of ensemble parameters.

    Parameters
    ----------
    ens:
        Ensemble object containing a set of parameter vectors
    normalize:
        flag indicating whether the returned Eigenvalues should be normalized
        with respect to the largest Eigenvalue
    only_separable_directions:
        return only separable directions according to
        cutoff_[absolute/relative]_separable
    cutoff_absolute_separable:
        Consider only eigenvalues of the covariance matrix above this cutoff
        (only applied when only_separable_directions is True)
    cutoff_relative_separable:
        Consider only eigenvalues of the covariance matrix above this cutoff,
        when rescaled with the largest eigenvalue
        (only applied when only_separable_directions is True)
    only_identifiable_directions:
        return only identifiable directions according to
        cutoff_[absolute/relative]_identifiable
    cutoff_absolute_identifiable:
        Consider only low eigenvalues of the covariance matrix with inverses
        above of this cutoff
        (only applied when only_identifiable_directions is True)
    cutoff_relative_identifiable:
        Consider only low eigenvalues of the covariance matrix when rescaled
        with the largest eigenvalue with inverses above of this cutoff
        (only applied when only_identifiable_directions is True)

    Returns
    -------
    eigenvalues:
        Eigenvalues of the covariance matrix
    eigenvectors:
        Eigenvectors of the covariance matrix
    """
    # check inputs
    if sum([only_identifiable_directions, only_separable_directions]) >= 2:
        raise AssertionError(
            "Specify either only identifiable or only separable directions."
        )

    covariance = get_covariance_matrix_parameters(ens)
    return get_spectral_decomposition_lowlevel(
        matrix=covariance,
        normalize=normalize,
        only_separable_directions=only_separable_directions,
        cutoff_absolute_separable=cutoff_absolute_separable,
        cutoff_relative_separable=cutoff_relative_separable,
        only_identifiable_directions=only_identifiable_directions,
        cutoff_absolute_identifiable=cutoff_absolute_identifiable,
        cutoff_relative_identifiable=cutoff_relative_identifiable,
    )


def get_spectral_decomposition_predictions(
    ens: Ensemble,
    normalize: bool = False,
    only_separable_directions: bool = False,
    cutoff_absolute_separable: float = 1e-16,
    cutoff_relative_separable: float = 1e-16,
    only_identifiable_directions: bool = False,
    cutoff_absolute_identifiable: float = 1e-16,
    cutoff_relative_identifiable: float = 1e-16,
) -> Tuple[np.ndarray, np.ndarray]:
    """
    Compute the spectral decomposition of ensemble predictions.

    Parameters
    ----------
    ens:
        Ensemble object containing a set of parameter vectors and a set of
        predictions or EnsemblePrediction object containing only predictions
    normalize:
        flag indicating whether the returned Eigenvalues should be normalized
        with respect to the largest Eigenvalue
    only_separable_directions:
        return only separable directions according to
        cutoff_[absolute/relative]_separable
    cutoff_absolute_separable:
        Consider only eigenvalues of the covariance matrix above this cutoff
        (only applied when only_separable_directions is True)
    cutoff_relative_separable:
        Consider only eigenvalues of the covariance matrix above this cutoff,
        when rescaled with the largest eigenvalue
        (only applied when only_separable_directions is True)
    only_identifiable_directions:
        return only identifiable directions according to
        cutoff_[absolute/relative]_identifiable
    cutoff_absolute_identifiable:
        Consider only low eigenvalues of the covariance matrix with inverses
        above of this cutoff
        (only applied when only_identifiable_directions is True)
    cutoff_relative_identifiable:
        Consider only low eigenvalues of the covariance matrix when rescaled
        with the largest eigenvalue with inverses above of this cutoff
        (only applied when only_identifiable_directions is True)

    Returns
    -------
    eigenvalues:
        Eigenvalues of the covariance matrix
    eigenvectors:
        Eigenvectors of the covariance matrix
    """
    covariance = get_covariance_matrix_predictions(ens)
    return get_spectral_decomposition_lowlevel(
        matrix=covariance,
        normalize=normalize,
        only_separable_directions=only_separable_directions,
        cutoff_absolute_separable=cutoff_absolute_separable,
        cutoff_relative_separable=cutoff_relative_separable,
        only_identifiable_directions=only_identifiable_directions,
        cutoff_absolute_identifiable=cutoff_absolute_identifiable,
        cutoff_relative_identifiable=cutoff_relative_identifiable,
    )


def get_spectral_decomposition_lowlevel(
    matrix: np.ndarray,
    normalize: bool = False,
    only_separable_directions: bool = False,
    cutoff_absolute_separable: float = 1e-16,
    cutoff_relative_separable: float = 1e-16,
    only_identifiable_directions: bool = False,
    cutoff_absolute_identifiable: float = 1e-16,
    cutoff_relative_identifiable: float = 1e-16,
) -> Tuple[np.ndarray, np.ndarray]:
    """
    Compute the spectral decomposition of ensemble parameters or predictions.

    Parameters
    ----------
    matrix:
        symmetric matrix (typically a covariance matrix) of parameters or
        predictions
    normalize:
        flag indicating whether the returned Eigenvalues should be normalized
        with respect to the largest Eigenvalue
    only_separable_directions:
        return only separable directions according to
        cutoff_[absolute/relative]_separable
    cutoff_absolute_separable:
        Consider only eigenvalues of the covariance matrix above this cutoff
        (only applied when only_separable_directions is True)
    cutoff_relative_separable:
        Consider only eigenvalues of the covariance matrix above this cutoff,
        when rescaled with the largest eigenvalue
        (only applied when only_separable_directions is True)
    only_identifiable_directions:
        return only identifiable directions according to
        cutoff_[absolute/relative]_identifiable
    cutoff_absolute_identifiable:
        Consider only low eigenvalues of the covariance matrix with inverses
        above of this cutoff
        (only applied when only_identifiable_directions is True)
    cutoff_relative_identifiable:
        Consider only low eigenvalues of the covariance matrix when rescaled
        with the largest eigenvalue with inverses above of this cutoff
        (only applied when only_identifiable_directions is True)

    Returns
    -------
    eigenvalues:
        Eigenvalues of the covariance matrix
    eigenvectors:
        Eigenvectors of the covariance matrix
    """
    # get the eigenvalue decomposition
    eigenvalues, eigenvectors = np.linalg.eigh(matrix)

    # get a normalized version
    rel_eigenvalues = eigenvalues / np.max(eigenvalues)

    # If no filtering is wanted, we can return
    if not only_identifiable_directions and not only_separable_directions:
        # apply normalization
        if normalize:
            eigenvalues = rel_eigenvalues
        return eigenvalues, eigenvectors

    # Separable directions are wanted: an upper pass filtering is needed
    if only_separable_directions:
        if (
            cutoff_absolute_separable is not None
            and cutoff_relative_separable is not None
        ):
            above_cutoff = np.array(
                [
                    i_eig_abs > cutoff_absolute_separable
                    and i_eig_rel > cutoff_relative_separable
                    for i_eig_abs, i_eig_rel in zip(
                        eigenvalues, rel_eigenvalues
                    )
                ]
            )
        elif cutoff_absolute_separable is not None:
            above_cutoff = eigenvalues > cutoff_absolute_separable
        elif cutoff_relative_separable is not None:
            above_cutoff = rel_eigenvalues > cutoff_relative_separable
        else:
            raise Exception(
                'Need a lower cutoff (absolute or relative, '
                'e.g., 1e-16, to compute separable directions.'
            )

        # restrict to those above cutoff
        eigenvalues = eigenvalues[above_cutoff]
        eigenvectors = eigenvectors[:, above_cutoff]
        # apply normlization
        if normalize:
            eigenvalues = rel_eigenvalues[above_cutoff]
        return eigenvalues, eigenvectors

    # Identifiable directions are wanted: an filtering of the inverse
    # eigenvalues is needed (upper pass of inverse = lower pass of original)
    if (
        cutoff_absolute_identifiable is not None
        and cutoff_relative_identifiable is not None
    ):
        below_cutoff = np.array(
            [
                1 / i_eig_abs > cutoff_absolute_identifiable
                and 1 / i_eig_rel > cutoff_relative_identifiable
                for i_eig_abs, i_eig_rel in zip(eigenvalues, rel_eigenvalues)
            ]
        )
    elif cutoff_absolute_identifiable is not None:
        below_cutoff = 1 / eigenvalues > cutoff_absolute_identifiable
    elif cutoff_relative_identifiable is not None:
        below_cutoff = 1 / rel_eigenvalues > cutoff_relative_identifiable
    else:
        raise Exception(
            'Need an inverse upper cutoff (absolute or relative, '
            'e.g., 1e-16, to compute identifiable directions.'
        )

    # restrict to those below cutoff
    eigenvalues = eigenvalues[below_cutoff]
    eigenvectors = eigenvectors[:, below_cutoff]
    # apply normlization
    if normalize:
        eigenvalues = rel_eigenvalues[below_cutoff]
    return eigenvalues, eigenvectors



"""
Startpoint
==========

Methods for selecting points that can be used as startpoints
for multi-start optimization.
Startpoint methods can be implemented by deriving from
:class:`pypesto.startpoint.StartpointMethod`.
"""

from .base import (
    CheckedStartpoints,
    FunctionStartpoints,
    NoStartpoints,
    StartpointMethod,
    to_startpoint_method,
)
from .latin_hypercube import LatinHypercubeStartpoints, latin_hypercube
from .uniform import UniformStartpoints, uniform



"""EmceeSampler class."""

from __future__ import annotations

import logging
from typing import List, Union

import numpy as np

from ..problem import Problem
from ..result import McmcPtResult
from ..startpoint import UniformStartpoints, uniform
from .sampler import Sampler, SamplerImportError

logger = logging.getLogger(__name__)


class EmceeSampler(Sampler):
    """Use emcee for sampling.

    Wrapper around https://emcee.readthedocs.io/en/stable/, see there for
    details.
    """

    def __init__(
        self,
        nwalkers: int = 1,
        sampler_args: dict = None,
        run_args: dict = None,
    ):
        """
        Initialize sampler.

        Parameters
        ----------
        nwalkers:
            The number of walkers in the ensemble.
        sampler_args:
            Further keyword arguments that are passed on to
            ``emcee.EnsembleSampler.__init__``.
        run_args:
            Further keyword arguments that are passed on to
            ``emcee.EnsembleSampler.run_mcmc``.
        """
        # check dependencies
        try:
            import emcee
        except ImportError:
            raise SamplerImportError("emcee")

        super().__init__()
        self.nwalkers: int = nwalkers

        if sampler_args is None:
            sampler_args = {}
        self.sampler_args: dict = sampler_args

        if run_args is None:
            run_args = {}
        self.run_args: dict = run_args

        # set in initialize
        self.problem: Union[Problem, None] = None
        self.sampler: Union[emcee.EnsembleSampler, None] = None
        self.state: Union[emcee.State, None] = None

    def get_epsilon_ball_initial_state(
        self,
        center: np.ndarray,
        problem: Problem,
        epsilon: float = 1e-3,
    ):
        """Get walker initial positions as samples from an epsilon ball.

        The ball is scaled in each direction according to the magnitude of the
        center in that direction.

        It is assumed that, because vectors are generated near a good point,
        all generated vectors are evaluable, so evaluability is not checked.

        Points that are generated outside the problem bounds will get shifted
        to lie on the edge of the problem bounds.

        Parameters
        ----------
        center:
            The center of the epsilon ball. The dimension should match the full
            dimension of the pyPESTO problem. This will be returned as the
            first position.
        problem:
            The pyPESTO problem.
        epsilon:
            The relative radius of the ball. e.g., if `epsilon=0.5`
            and the center of the first dimension is at 100, then the upper
            and lower bounds of the epsilon ball in the first dimension will
            be 150 and 50, respectively.
        """
        # Epsilon ball
        lb = center * (1 - epsilon)
        ub = center * (1 + epsilon)

        # Adjust bounds to satisfy problem bounds
        lb[lb < problem.lb] = problem.lb[lb < problem.lb]
        ub[ub > problem.ub] = problem.ub[ub > problem.ub]

        # Sample initial positions
        initial_state_after_first = uniform(
            n_starts=self.nwalkers - 1,
            lb=lb,
            ub=ub,
        )

        # Include `center` in initial positions
        initial_state = np.row_stack(
            (
                center,
                initial_state_after_first,
            )
        )

        return initial_state

    def initialize(
        self,
        problem: Problem,
        x0: Union[np.ndarray, List[np.ndarray]],
    ) -> None:
        """Initialize the sampler.

        It is recommended to initialize walkers

        Parameters
        ----------
        x0:
            The "a priori preferred position". e.g., an optimized parameter
            vector. https://emcee.readthedocs.io/en/stable/user/faq/
            The position of the first walker will be this, the remaining
            walkers will be assigned positions uniformly in a smaller ball
            around this vector.
            Alternatively, a set of vectors can be provided, which will be used
            to initialize walkers. In this case, any remaining walkers will be
            initialized at points sampled uniformly within the problem bounds.
        """
        import emcee

        self.problem = problem

        # extract for pickling efficiency
        objective = self.problem.objective
        lb = self.problem.lb
        ub = self.problem.ub

        # parameter dimenstion
        ndim = len(self.problem.x_free_indices)

        def log_prob(x):
            """Log-probability density function."""
            # check if parameter lies within bounds
            if any(x < lb) or any(x > ub):
                return -np.inf
            # invert sign
            return -1.0 * objective(x)

        # initialize sampler
        self.sampler = emcee.EnsembleSampler(
            nwalkers=self.nwalkers,
            ndim=ndim,
            log_prob_fn=log_prob,
            **self.sampler_args,
        )

        # assign startpoints
        if self.state is None:
            if x0.ndim > 1 and len(x0.shape[0]) > 1:
                logger.warning(
                    "More than a single vector was provided to initialize the "
                    "walker positions. If these vectors do not exist in a "
                    "small ball around a high-probability position (e.g. "
                    "optimized vector) then sampling may be inefficient (see "
                    "emcee FAQ: "
                    "https://emcee.readthedocs.io/en/stable/user/faq/ )."
                )
                #  extract x0
                x0 = np.asarray(x0)
                if x0.ndim == 1:
                    x0 = [x0]
                x0 = np.array([problem.get_full_vector(x) for x in x0])
                x_guesses_full0 = problem.x_guesses_full
                #  add x0 to guesses
                problem.set_x_guesses(
                    np.row_stack(
                        (
                            x0,
                            problem.x_guesses_full,
                        )
                    )
                )
                #  sample start points
                initial_state = UniformStartpoints(
                    use_guesses=True,
                    check_fval=True,
                    check_grad=False,
                )(
                    n_starts=self.nwalkers,
                    problem=problem,
                )
                #  restore original guesses
                problem.set_x_guesses(x_guesses_full0)
            else:
                initial_state = self.get_epsilon_ball_initial_state(
                    center=x0,
                    problem=problem,
                )

            self.state = initial_state

    def sample(self, n_samples: int, beta: float = 1.0) -> None:
        """Return the most recent sample state."""
        self.state = self.sampler.run_mcmc(
            initial_state=self.state,
            nsteps=n_samples,
            **self.run_args,
        )

    def get_samples(self) -> McmcPtResult:
        """Get the samples into the fitting pypesto format."""
        # all walkers are concatenated, yielding a flat array
        trace_x = np.array([self.sampler.get_chain(flat=True)])
        trace_neglogpost = np.array([-self.sampler.get_log_prob(flat=True)])
        # the sampler does not know priors
        trace_neglogprior = np.full(trace_neglogpost.shape, np.nan)
        # the walkers all run on temperature 1
        betas = np.array([1.0])

        result = McmcPtResult(
            trace_x=trace_x,
            trace_neglogpost=trace_neglogpost,
            trace_neglogprior=trace_neglogprior,
            betas=betas,
        )

        return result



"""Process a model selection :class:`ModelProblem` after calibration."""
from pathlib import Path
from typing import List

import matplotlib.pyplot as plt
import numpy as np
from petab_select.constants import ESTIMATE, TYPE_PATH, Criterion

from .. import store, visualize
from ..C import TYPE_POSTPROCESSOR
from .model_problem import ModelProblem


def multi_postprocessor(
    problem: ModelProblem,
    postprocessors: List[TYPE_POSTPROCESSOR] = None,
):
    """Combine multiple postprocessors into a single postprocessor.

    See `save_postprocessor` for usage hints.

    Parameters
    ----------
    problem:
        A model selection :class:`ModelProblem` that has been optimized.
    postprocessors:
        A list of postprocessors, which will be sequentially applied to the
        optimized model `problem`.
        The location where results will be stored.
    """
    for postprocessor in postprocessors:
        postprocessor(problem)


def waterfall_plot_postprocessor(
    problem: ModelProblem,
    output_path: TYPE_PATH = ".",
):
    """Produce a waterfall plot.

    See `save_postprocessor` for usage hints and argument documentation.
    """
    visualize.waterfall(problem.minimize_result)
    plot_output_path = Path(output_path) / (problem.model.model_hash + ".png")
    plt.savefig(str(plot_output_path))


def save_postprocessor(
    problem: ModelProblem,
    output_path: TYPE_PATH = ".",
    use_model_hash: bool = False,
):
    """Save the parameter estimation result.

    When used, first set the output folder for results, e.g. with
    `functools.partial`. This is because postprocessors should take only a
    single parameter: an optimized model.
    .. code-block:: python
       from functools import partial
       output_path = 'results'
       pp = partial(save_postprocessor, output_path=output_path)
       selector = pypesto.select.ModelSelector(
           problem=problem,
           model_postprocessor=pp,
       )

    Parameters
    ----------
    problem:
        A model selection :class:`ModelProblem` that has been optimized.
    output_path:
        The location where output will be stored.
    use_model_hash:
        Whether the filename should use the model hash. Defaults to `False`,
        in which case the model ID is used instead.
    """
    stem = problem.model.model_id
    if use_model_hash:
        stem = problem.model.get_hash()
    store.write_result(
        problem.minimize_result,
        Path(output_path) / (stem + ".hdf5"),
    )


def model_id_binary_postprocessor(problem: ModelProblem):
    """Change a PEtab Select model ID to a binary string.

    Changes the model ID in-place to be a string like `M_ijk`, where
    `i`, `j`, `k`, etc. are `1` if the parameter in that position is estimated,
    or `0` if the parameter is fixed.

    To ensure that other postprocessors (e.g. `report_postprocessor`) use this
    new model ID, when in use with a `multi_postprocessor`, ensure this is
    before the other postprocessors in the `postprocessors` argument of
    `multi_postprocessor`.

    Parameters
    ----------
    problem:
        A model selection :class:`ModelProblem` that has been optimized.
    """
    model_id = "M_"
    for parameter_value in problem.model.parameters.values():
        model_id += "1" if parameter_value == ESTIMATE else "0"
    problem.model.model_id = model_id


def report_postprocessor(
    problem: ModelProblem,
    output_filepath: TYPE_PATH,
    criteria: List[Criterion] = None,
):
    """Create a TSV table of model selection results.

    Parameters
    ----------
    problem:
        A model selection :class:`ModelProblem` that has been optimized.
    output_filepath:
        The file path where the report will be saved.
    criteria:
        The criteria that will be in the report. Defaults to nllh, AIC, AICc,
        and BIC.
    """
    output_filepath = Path(output_filepath)
    write_header = False
    # Only write the header if the file doesn't yet exist or is empty.
    if not output_filepath.exists() or output_filepath.stat().st_size == 0:
        write_header = True
    if criteria is None:
        criteria = [
            Criterion.NLLH,
            Criterion.AIC,
            Criterion.AICC,
            Criterion.BIC,
        ]

    start_optimization_times = problem.minimize_result.optimize_result.time

    header = []
    row = []

    header.append('model_id')
    row.append(problem.model.model_id)

    header.append('total_time')
    row.append(str(sum(start_optimization_times)))

    for criterion in criteria:
        header.append(criterion.value)
        row.append(str(problem.model.get_criterion(criterion)))

    # Arbitrary convergence criterion
    header.append('n_converged')
    row.append(
        str(
            (
                np.array(problem.minimize_result.optimize_result.fval)
                < (problem.minimize_result.optimize_result.list[0].fval + 0.1)
            ).sum()
        )
    )

    for start_index, start_optimization_time in enumerate(
        start_optimization_times
    ):
        header.append(f'start_time_{start_index}')
        row.append(str(start_optimization_time))

    with open(output_filepath, 'a+') as f:
        if write_header:
            f.write('\t'.join(header) + '\n')
        f.write('\t'.join(row) + '\n')



"""In-memory history."""

import time
from typing import Any, Dict, Sequence, Tuple, Union

import numpy as np

from ..C import FVAL, GRAD, HESS, RES, SRES, TIME, ModeType, X
from .base import (
    CountHistoryBase,
    HistoryBase,
    add_fun_from_res,
    reduce_result_via_options,
)
from .options import HistoryOptions
from .util import MaybeArray, ResultDict, trace_wrap


class MemoryHistory(CountHistoryBase):
    """
    Class for optimization history stored in memory.

    Tracks number of function evaluations and keeps an in-memory
    trace of function evaluations.

    Parameters
    ----------
    options:
        History options.
    """

    def __init__(self, options: Union[HistoryOptions, Dict] = None):
        super().__init__(options=options)
        self._trace: Dict[str, Any] = {key: [] for key in HistoryBase.ALL_KEYS}

    def update(
        self,
        x: np.ndarray,
        sensi_orders: Tuple[int, ...],
        mode: ModeType,
        result: ResultDict,
    ) -> None:
        """See `History` docstring."""
        super().update(x, sensi_orders, mode, result)
        self._update_trace(x, mode, result)

    def _update_trace(self, x, mode, result):
        """Update internal trace representation."""
        # calculating function values from residuals
        #  and reduce via requested history options
        result: dict = reduce_result_via_options(
            add_fun_from_res(result), self.options
        )

        result[X] = x

        used_time = time.time() - self._start_time
        result[TIME] = used_time

        for key in HistoryBase.ALL_KEYS:
            self._trace[key].append(result[key])

    def __len__(self) -> int:
        """Define length of history object."""
        return len(self._trace[TIME])

    @trace_wrap
    def get_x_trace(
        self,
        ix: Union[int, Sequence[int], None] = None,
        trim: bool = False,
    ) -> Union[Sequence[np.ndarray], np.ndarray]:
        """See `HistoryBase` docstring."""
        return [self._trace[X][i] for i in ix]

    @trace_wrap
    def get_fval_trace(
        self,
        ix: Union[int, Sequence[int], None] = None,
        trim: bool = False,
    ) -> Union[Sequence[float], float]:
        """See `HistoryBase` docstring."""
        return [self._trace[FVAL][i] for i in ix]

    @trace_wrap
    def get_grad_trace(
        self,
        ix: Union[int, Sequence[int], None] = None,
        trim: bool = False,
    ) -> Union[Sequence[MaybeArray], MaybeArray]:
        """See `HistoryBase` docstring."""
        return [self._trace[GRAD][i] for i in ix]

    @trace_wrap
    def get_hess_trace(
        self,
        ix: Union[int, Sequence[int], None] = None,
        trim: bool = False,
    ) -> Union[Sequence[MaybeArray], MaybeArray]:
        """See `HistoryBase` docstring."""
        return [self._trace[HESS][i] for i in ix]

    @trace_wrap
    def get_res_trace(
        self,
        ix: Union[int, Sequence[int], None] = None,
        trim: bool = False,
    ) -> Union[Sequence[MaybeArray], MaybeArray]:
        """See `HistoryBase` docstring."""
        return [self._trace[RES][i] for i in ix]

    @trace_wrap
    def get_sres_trace(
        self,
        ix: Union[int, Sequence[int], None] = None,
        trim: bool = False,
    ) -> Union[Sequence[MaybeArray], MaybeArray]:
        """See `HistoryBase` docstring."""
        return [self._trace[SRES][i] for i in ix]

    @trace_wrap
    def get_time_trace(
        self,
        ix: Union[int, Sequence[int], None] = None,
        trim: bool = False,
    ) -> Union[Sequence[float], float]:
        """See `HistoryBase` docstring."""
        return [self._trace[TIME][i] for i in ix]



import logging
from typing import Sequence, Tuple

from ..C import ModeType
from ..engine import Task

logger = logging.getLogger(__name__)


class PredictorTask(Task):
    """Perform a single prediction with :class:`pypesto.engine.Task`.

    Designed for use with :class:`pypesto.ensemble.Ensemble`.

    Attributes
    ----------
    predictor:
        The predictor to use.
    x:
        The parameter vector to compute predictions with.
    sensi_orders:
        Specifies which sensitivities to compute, e.g. (0,1) -> fval, grad.
    mode:
        Whether to compute function values or residuals.
    id:
        The input ID.
    """

    def __init__(
        self,
        predictor: 'pypesto.predict.Predictor',  # noqa: F821
        x: Sequence[float],
        sensi_orders: Tuple[int, ...],
        mode: ModeType,
        id: str,
    ):
        super().__init__()
        self.predictor = predictor
        self.x = x
        self.sensi_orders = sensi_orders
        self.mode = mode
        self.id = id

    def execute(self) -> 'pypesto.predict.PredictionResult':  # noqa: F821
        """Execute and return the prediction."""
        logger.info(f"Executing task {self.id}.")
        prediction = self.predictor(self.x, self.sensi_orders, self.mode)
        return prediction



"""Visualization routines for model selection with pyPESTO."""
from typing import Dict, List, Tuple

import matplotlib.pyplot as plt
import networkx as nx
from petab_select import Model
from petab_select.constants import VIRTUAL_INITIAL_MODEL, Criterion

from .. import select as pypesto_select

# TODO move methods to petab_select
RELATIVE_LABEL_FONTSIZE = -2


def default_label_maker(model: Model) -> str:
    """Create a model label, for plotting."""
    return model.model_hash[:4]


# FIXME supply the problem to automatically detect the correct criterion?
def plot_selected_models(
    selected_models: List[Model],
    criterion: str = Criterion.AIC,
    relative: str = True,
    fz: int = 14,
    size: Tuple[float, float] = (5, 4),
    labels: Dict[str, str] = None,
    ax: plt.Axes = None,
) -> plt.Axes:
    """Plot criterion for calibrated models.

    Parameters
    ----------
    selected_models:
        First result in tuple returned by `ModelSelector.select()`.
    criterion:
        Key of criterion value in `selected_models` to be plotted.
    relative:
        If `True`, criterion values are plotted relative to the lowest
        criterion value. TODO is the lowest value, always the best? May not
        be for different criterion.
    fz:
        fontsize
    size:
        Figure size in inches.
    labels:
        A dictionary of model labels, where keys are model hashes, and
        values are model labels, for plotting. If a model label is not
        provided, it will be generated from its model ID.
    ax:
        The axis to use for plotting.

    Returns
    -------
    matplotlib.pyplot.Axes
        The plot axis.
    """
    zero = 0
    if relative:
        zero = selected_models[-1].get_criterion(criterion)

    if labels is None:
        labels = {}

    # FIGURE
    if ax is None:
        _, ax = plt.subplots(figsize=size)
    linewidth = 3

    selected_models = [
        model for model in selected_models if model != VIRTUAL_INITIAL_MODEL
    ]

    criterion_values = {
        labels.get(
            model.get_hash(), default_label_maker(model)
        ): model.get_criterion(criterion)
        - zero
        for model in selected_models
    }

    ax.plot(
        criterion_values.keys(),
        criterion_values.values(),
        linewidth=linewidth,
        color='lightgrey',
        # edgecolor='k'
    )

    ax.get_xticks()
    ax.set_xticks(list(range(len(criterion_values))))
    ax.set_ylabel(
        criterion + ('(relative)' if relative else '(absolute)'), fontsize=fz
    )
    # could change to compared_model_ids, if all models are plotted
    ax.set_xticklabels(
        criterion_values.keys(),
        fontsize=fz + RELATIVE_LABEL_FONTSIZE,
    )
    for tick in ax.yaxis.get_major_ticks():
        tick.label1.set_fontsize(fz + RELATIVE_LABEL_FONTSIZE)
    ytl = ax.get_yticks()
    ax.set_ylim([min(ytl), max(ytl)])
    # removing top and right borders
    ax.spines['top'].set_visible(False)
    ax.spines['right'].set_visible(False)
    return ax


def plot_history_digraph(*args, **kwargs):
    """Ignore me. Renamed to `plot_calibrated_models_digraph`."""
    raise NotImplementedError(
        "This method is now `plot_calibrated_models_digraph`."
    )


def plot_calibrated_models_digraph(
    problem: pypesto_select.Problem,
    calibrated_models: Dict[str, Model] = None,
    criterion: Criterion = None,
    optimal_distance: float = 1,
    relative: bool = True,
    options: Dict = None,
    labels: Dict[str, str] = None,
    ax: plt.Axes = None,
) -> plt.Axes:
    """Plot all calibrated models in the model space, as a directed graph.

    TODO replace magic numbers with options/constants

    Parameters
    ----------
    problem:
        The pyPESTO Select problem.
    calibrated_models:
        The models calibrated during model selection, in the format of
        `pypesto.select.Problem.calibrated_models`.
    criterion:
        The criterion.
    optimal_distance:
        See docs for argument `k` in `networkx.spring_layout`.
    relative:
        If `True`, criterion values are offset by the minimum criterion
        value.
    options:
        Additional keyword arguments for `networkx.draw_networkx`.
    labels:
        A dictionary of model labels, where keys are model hashes, and
        values are model labels, for plotting. If a model label is not
        provided, it will be generated from its model ID.
    ax:
        The axis to use for plotting.

    Returns
    -------
    matplotlib.pyplot.Axes
        The plot axis.
    """
    if criterion is None:
        criterion = problem.petab_select_problem.criterion
    if calibrated_models is None:
        calibrated_models = problem.calibrated_models
    if labels is None:
        labels = {}

    G = nx.DiGraph()
    edges = []
    for model in calibrated_models.values():
        predecessor_model_hash = model.predecessor_model_hash
        if predecessor_model_hash is not None:
            from_ = labels.get(predecessor_model_hash, predecessor_model_hash)
            # may only not be the case for
            # COMPARED_MODEL_ID == INITIAL_VIRTUAL_MODEL
            if predecessor_model_hash in calibrated_models:
                predecessor_model = calibrated_models[predecessor_model_hash]
                from_ = labels.get(
                    predecessor_model.get_hash(),
                    default_label_maker(predecessor_model),
                )
        else:
            raise NotImplementedError(
                'Plots for models with `None` as their predecessor model are '
                'not yet implemented.'
            )
            from_ = 'None'
        to = labels.get(model.get_hash(), default_label_maker(model))
        edges.append((from_, to))

    G.add_edges_from(edges)
    default_options = {
        'node_color': 'lightgrey',
        'arrowstyle': '-|>',
        'node_shape': 's',
        'node_size': 2500,
    }
    if options is not None:
        default_options.update(options)

    if ax is None:
        _, ax = plt.subplots(figsize=(12, 12))

    pos = nx.spring_layout(G, k=optimal_distance, iterations=20)
    nx.draw_networkx(G, pos, ax=ax, **default_options)

    return ax



import copy
from typing import Any, Dict, List

import numpy as np

from ..C import InnerParameterType
from ..objective import Objective
from ..optimize import minimize
from ..problem import Problem
from .problem import InnerProblem, scale_value_dict
from .util import (
    apply_offset,
    apply_scaling,
    apply_sigma,
    compute_nllh,
    compute_optimal_offset,
    compute_optimal_offset_coupled,
    compute_optimal_scaling,
    compute_optimal_sigma,
)


class InnerSolver:
    """Solver for an inner optimization problem."""

    def initialize(self):
        """
        (Re-)initialize the solver.

        Default: Do nothing.
        """

    def solve(
        self,
        problem: InnerProblem,
        sim: List[np.ndarray],
        sigma: List[np.ndarray],
        scaled: bool,
    ) -> Dict[str, float]:
        """Solve the subproblem.

        Parameters
        ----------
        problem:
            The inner problem to solve.
        sim:
            List of model output matrices, as provided in AMICI's
            ``ReturnData.y``. Same order as simulations in the
            PEtab problem.
        sigma:
            List of sigma matrices from the model, as provided in AMICI's
            ``ReturnData.sigmay``. Same order as simulations in the
            PEtab problem.
        scaled:
            Whether to scale the results to the parameter scale specified in
            ``problem``.
        """


class AnalyticalInnerSolver(InnerSolver):
    """Solve the inner subproblem analytically.

    Currently, supports offset and scaling parameters (coupled or not), and
    sigmas for additive Gaussian noise.
    """

    def solve(
        self,
        problem: InnerProblem,
        sim: List[np.ndarray],
        sigma: List[np.ndarray],
        scaled: bool,
    ) -> Dict[str, float]:
        """Solve the subproblem analytically.

        Parameters
        ----------
        problem:
            The inner problem to solve.
        sim:
            List of model output matrices, as provided in AMICI's
            ``ReturnData.y``. Same order as simulations in the
            PEtab problem.
        sigma:
            List of sigma matrices from the model, as provided in AMICI's
            ``ReturnData.sigmay``. Same order as simulations in the
            PEtab problem.
        scaled:
            Whether to scale the results to the parameter scale specified in
            ``problem``.
        """
        x_opt = {}

        data = copy.deepcopy(problem.data)

        # compute optimal offsets
        for x in problem.get_xs_for_type(InnerParameterType.OFFSET):
            if x.coupled:
                x_opt[x.inner_parameter_id] = compute_optimal_offset_coupled(
                    data=data, sim=sim, sigma=sigma, mask=x.ixs
                )
            else:
                x_opt[x.inner_parameter_id] = compute_optimal_offset(
                    data=data, sim=sim, sigma=sigma, mask=x.ixs
                )
        # apply offsets
        for x in problem.get_xs_for_type(InnerParameterType.OFFSET):
            apply_offset(
                offset_value=x_opt[x.inner_parameter_id], data=data, mask=x.ixs
            )

        # compute optimal scalings
        for x in problem.get_xs_for_type(InnerParameterType.SCALING):
            x_opt[x.inner_parameter_id] = compute_optimal_scaling(
                data=data, sim=sim, sigma=sigma, mask=x.ixs
            )
        # apply scalings
        for x in problem.get_xs_for_type(InnerParameterType.SCALING):
            apply_scaling(
                scaling_value=x_opt[x.inner_parameter_id], sim=sim, mask=x.ixs
            )

        # compute optimal sigmas
        for x in problem.get_xs_for_type(InnerParameterType.SIGMA):
            x_opt[x.inner_parameter_id] = compute_optimal_sigma(
                data=data, sim=sim, mask=x.ixs
            )
        # apply sigmas
        for x in problem.get_xs_for_type(InnerParameterType.SIGMA):
            apply_sigma(
                sigma_value=x_opt[x.inner_parameter_id],
                sigma=sigma,
                mask=x.ixs,
            )

        # scale
        if scaled:
            x_opt = scale_value_dict(x_opt, problem)
        return x_opt


class NumericalInnerSolver(InnerSolver):
    """Solve the inner subproblem numerically.

    Advantage: The structure of the subproblem does not matter like, at all.
    Disadvantage: Slower.

    Special features: We cache the best parameters, which substantially
    speeds things up.

    Attributes
    ----------
    minimize_kwargs:
        Passed to the `pypesto.optimize.minimize` call.
    n_cached:
        Number of optimized parameter vectors to save.
    problem_kwargs:
        Passed to the `pypesto.Problem` constructor.
    x_guesses:
        Cached optimized parameter vectors, supplied as guesses to the next
        `solve` call.
    """

    def __init__(
        self,
        minimize_kwargs: Dict[str, Any] = None,
        n_cached: int = 1,
        problem_kwargs: Dict[str, Any] = None,
    ):
        self.minimize_kwargs = minimize_kwargs
        if self.minimize_kwargs is None:
            self.minimize_kwargs = {}
        self.n_cached = n_cached
        self.problem_kwargs = problem_kwargs
        if self.problem_kwargs is None:
            self.problem_kwargs = {}

        self.minimize_kwargs['n_starts'] = self.minimize_kwargs.get(
            'n_starts', 1
        )
        self.minimize_kwargs['progress_bar'] = self.minimize_kwargs.get(
            'progress_bar', False
        )

        self.x_guesses = None
        self.dummy_lb = -1e20
        self.dummy_ub = +1e20

    def initialize(self):
        """(Re-)initialize the solver."""
        self.x_guesses = None

    def solve(
        self,
        problem: InnerProblem,
        sim: List[np.ndarray],
        sigma: List[np.ndarray],
        scaled: bool,
    ) -> Dict[str, float]:
        """Solve the subproblem numerically.

        Parameters
        ----------
        problem:
            The inner problem to solve.
        sim:
            List of model output matrices, as provided in AMICI's
            ``ReturnData.y``. Same order as simulations in the
            PEtab problem.
        sigma:
            List of sigma matrices from the model, as provided in AMICI's
            ``ReturnData.sigmay``. Same order as simulations in the
            PEtab problem.
        scale:
            Whether to scale the results to the parameter scale specified in
            ``problem``.
        """
        pars = problem.xs.values()
        # We currently cannot handle constraints on inner parameters correctly,
        # and would have to assume [-inf, inf]. However, this may not be
        # supported by all inner optimizers, so we go for some (arbitrary)
        # large value.
        lb = np.array(
            [
                0
                if x.inner_parameter_type == InnerParameterType.SIGMA
                else self.dummy_lb
                for x in pars
            ]
        )

        ub = np.full(shape=len(pars), fill_value=self.dummy_ub)

        x_names = [x.inner_parameter_id for x in pars]
        data = problem.data

        # objective function
        def fun(x):
            _sim = copy.deepcopy(sim)
            _sigma = copy.deepcopy(sigma)
            _data = copy.deepcopy(data)
            for x_val, par in zip(x, pars):
                mask = par.ixs
                if par.inner_parameter_type == InnerParameterType.OFFSET:
                    apply_offset(x_val, _data, mask)
                elif par.inner_parameter_type == InnerParameterType.SCALING:
                    apply_scaling(x_val, _sim, mask)
                elif par.inner_parameter_type == InnerParameterType.SIGMA:
                    apply_sigma(x_val, _sigma, mask)
                else:
                    raise ValueError(
                        "Can't handle parameter type "
                        f"`{par.inner_parameter_type}`."
                    )

            return compute_nllh(_data, _sim, _sigma)

        # TODO gradient
        objective = Objective(fun)

        # optimization problem
        pypesto_problem = Problem(
            objective, lb=lb, ub=ub, x_names=x_names, **self.problem_kwargs
        )

        if self.x_guesses is not None:
            pypesto_problem.set_x_guesses(
                self.x_guesses[:, pypesto_problem.x_free_indices]
            )
        else:
            pypesto_problem.set_x_guesses(
                [list(problem.get_dummy_values(scaled=False).values())]
            )

        # perform the actual optimization
        result = minimize(pypesto_problem, **self.minimize_kwargs)

        best_par = result.optimize_result.list[0]['x']

        if (np.isclose(best_par, lb) | np.isclose(best_par, ub)).any():
            raise RuntimeError(
                "Active bounds in inner problem optimization. This can result "
                "in incorrect gradient computation for the outer parameters."
            )

        x_opt = dict(zip(pypesto_problem.x_names, best_par))

        # cache
        self.x_guesses = np.array(
            [
                entry['x']
                for entry in result.optimize_result.list[: self.n_cached]
            ]
        )

        # scale
        if scaled:
            x_opt = scale_value_dict(x_opt, problem)

        return x_opt



import logging
from time import process_time
from typing import Callable, List, Optional, Union

import numpy as np

from ..problem import Problem
from ..result import Result
from ..store import autosave
from .adaptive_metropolis import AdaptiveMetropolisSampler
from .sampler import Sampler
from .util import bound_n_samples_from_env

logger = logging.getLogger(__name__)


def sample(
    problem: Problem,
    n_samples: Optional[int],
    sampler: Sampler = None,
    x0: Union[np.ndarray, List[np.ndarray]] = None,
    result: Result = None,
    filename: Union[str, Callable, None] = None,
    overwrite: bool = False,
) -> Result:
    """
    Call to do parameter sampling.

    Parameters
    ----------
    problem:
        The problem to be solved. If None is provided, a
        :class:`pypesto.AdaptiveMetropolisSampler` is used.
    n_samples:
        Number of samples to generate. `None` can be used if the sampler does
        not use `n_samples`.
    sampler:
        The sampler to perform the actual sampling.
    x0:
        Initial parameter for the Markov chain. If None, the best parameter
        found in optimization is used. Note that some samplers require an
        initial parameter, some may ignore it. x0 can also be a list,
        to have separate starting points for parallel tempering chains.
    result:
        A result to write to. If None provided, one is created from the
        problem.
    filename:
        Name of the hdf5 file, where the result will be saved. Default is
        None, which deactivates automatic saving. If set to
        "Auto" it will automatically generate a file named
        `year_month_day_profiling_result.hdf5`.
        Optionally a method, see docs for `pypesto.store.auto.autosave`.
    overwrite:
        Whether to overwrite `result/sampling` in the autosave file
        if it already exists.

    Returns
    -------
    result:
        A result with filled in sample_options part.
    """
    # prepare result object
    if result is None:
        result = Result(problem)

    # number of samples
    if n_samples is not None:
        n_samples = bound_n_samples_from_env(n_samples)

    # try to find initial parameters
    if x0 is None:
        result.optimize_result.sort()
        if len(result.optimize_result.list) > 0:
            x0 = problem.get_reduced_vector(
                result.optimize_result.list[0]['x']
            )
        # TODO multiple x0 for PT, #269

    # set sampler
    if sampler is None:
        sampler = AdaptiveMetropolisSampler()

    # initialize sampler to problem
    sampler.initialize(problem=problem, x0=x0)

    # perform the sampling and track time
    t_start = process_time()
    sampler.sample(n_samples=n_samples)
    t_elapsed = process_time() - t_start
    logger.info("Elapsed time: " + str(t_elapsed))

    # extract results
    sampler_result = sampler.get_samples()

    # record time
    sampler_result.time = t_elapsed

    # record results
    result.sample_result = sampler_result

    autosave(
        filename=filename,
        result=result,
        store_type="sample",
        overwrite=overwrite,
    )

    return result



"""Ensemble utilities."""

import os
from pathlib import Path
from typing import Callable, Literal, Sequence, Union

import h5py
import numpy as np
import pandas as pd

from ..C import (
    LOWER_BOUND,
    OPTIMIZE,
    OUTPUT,
    OUTPUT_IDS,
    OUTPUT_SIGMAY,
    OUTPUT_WEIGHT,
    PREDICTION_ID,
    PREDICTION_RESULTS,
    SAMPLE,
    SUMMARY,
    TIMEPOINTS,
    UPPER_BOUND,
    X_NAMES,
    EnsembleType,
)
from ..result import PredictionConditionResult, PredictionResult
from ..store import read_result, write_array
from .ensemble import Ensemble, EnsemblePrediction


def read_from_csv(
    path: str,
    sep: str = '\t',
    index_col: int = 0,
    headline_parser: Callable = None,
    ensemble_type: EnsembleType = None,
    lower_bound: np.ndarray = None,
    upper_bound: np.ndarray = None,
):
    """
    Create an ensemble from a csv file.

    Parameters
    ----------
    path:
        path to csv file to read in parameter ensemble
    sep:
        separator in csv file
    index_col:
        index column in csv file
    headline_parser:
        A function which reads in the headline of the csv file and converts it
        into vector_tags (see constructor of Ensemble for more details)
    ensemble_type:
        Ensemble type: representative sample or random ensemble
    lower_bound:
        array of potential lower bounds for the parameters
    upper_bound:
        array of potential upper bounds for the parameters

    Returns
    -------
    result:
        Ensemble object of parameter vectors
    """
    # get the data from the csv
    ensemble_df = pd.read_csv(path, sep=sep, index_col=index_col)

    # set the type of the ensemble
    if ensemble_type is None:
        ensemble_type = EnsembleType.ensemble

    return read_from_df(
        dataframe=ensemble_df,
        headline_parser=headline_parser,
        ensemble_type=ensemble_type,
        lower_bound=lower_bound,
        upper_bound=upper_bound,
    )


def read_ensemble_from_hdf5(
    filename: str,
    input_type: Literal['optimize', 'sample'] = OPTIMIZE,
    remove_burn_in: bool = True,
    chain_slice: slice = None,
    cutoff: float = np.inf,
    max_size: int = np.inf,
):
    """
    Create an ensemble from an HDF5 storage file.

    Parameters
    ----------
    filename:
        Name or path of the HDF5 file.
    input_type:
        Which type of ensemble to create. From History, from
        Optimization or from Sample.

    Returns
    -------
    ensemble:
        Ensemble object of parameter vectors
    """
    # TODO: add option HISTORY. Need to fix
    #  reading history from hdf5.
    if input_type == OPTIMIZE:
        result = read_result(filename=filename, optimize=True)
        return Ensemble.from_optimization_endpoints(
            result=result, rel_cutoff=cutoff, max_size=max_size
        )
    elif input_type == SAMPLE:
        result = read_result(filename=filename, sample=True)
        return Ensemble.from_sample(
            result=result,
            remove_burn_in=remove_burn_in,
            chain_slice=chain_slice,
        )
    else:
        raise ValueError(
            'The type you provided was neither '
            f'"{SAMPLE}" nor "{OPTIMIZE}". Those are '
            'currently the only supported types. '
            'Please choose one of them.'
        )


def read_from_df(
    dataframe: pd.DataFrame,
    headline_parser: Callable = None,
    ensemble_type: EnsembleType = None,
    lower_bound: np.ndarray = None,
    upper_bound: np.ndarray = None,
):
    """
    Create an ensemble from a csv file.

    Parameters
    ----------
    dataframe:
        pandas.DataFrame to read in parameter ensemble
    headline_parser:
        A function which reads in the headline of the csv file and converts it
        into vector_tags (see constructor of Ensemble for more details)
    ensemble_type:
        Ensemble type: representative sample or random ensemble
    lower_bound:
        array of potential lower bounds for the parameters
    upper_bound:
        array of potential upper bounds for the parameters

    Returns
    -------
    result:
        Ensemble object of parameter vectors
    """
    # if we have a parser to make vector_tags from column names, we use it
    vector_tags = None
    if headline_parser is not None:
        vector_tags = headline_parser(list(dataframe.columns))
    # set the type of the ensemble
    if ensemble_type is None:
        ensemble_type = EnsembleType.ensemble

    return Ensemble(
        x_vectors=dataframe.values,
        x_names=list(dataframe.index),
        vector_tags=vector_tags,
        ensemble_type=ensemble_type,
        lower_bound=lower_bound,
        upper_bound=upper_bound,
    )


def write_ensemble_prediction_to_h5(
    ensemble_prediction: EnsemblePrediction,
    output_file: str,
    base_path: str = None,
):
    """
    Write an `EnsemblePrediction` to hdf5.

    Parameters
    ----------
    ensemble_prediction:
        The prediciton to be saved.
    output_file:
        The filename of the hdf5 file.
    base_path:
        An optional filepath where the file should be saved to.
    """
    # parse base path
    base = Path('')
    if base_path is not None:
        base = Path(base_path)

    # open file
    with h5py.File(output_file, 'a') as f:
        # write prediction ID if available
        if ensemble_prediction.prediction_id is not None:
            f.create_dataset(
                os.path.join(base, PREDICTION_ID),
                data=ensemble_prediction.prediction_id,
            )

        # write lower bounds per condition, if available
        if ensemble_prediction.lower_bound is not None:
            if isinstance(ensemble_prediction.lower_bound[0], np.ndarray):
                lb_grp = f.require_group(LOWER_BOUND)
                for i_cond, lower_bounds in enumerate(
                    ensemble_prediction.lower_bound
                ):
                    condition_id = ensemble_prediction.prediction_results[
                        0
                    ].condition_ids[i_cond]
                    write_array(lb_grp, condition_id, lower_bounds)
            elif isinstance(ensemble_prediction.lower_bound[0], float):
                f.create_dataset(
                    LOWER_BOUND, data=ensemble_prediction.lower_bound
                )

        # write upper bounds per condition, if available
        if ensemble_prediction.upper_bound is not None:
            if isinstance(ensemble_prediction.upper_bound[0], np.ndarray):
                ub_grp = f.require_group(UPPER_BOUND)
                for i_cond, upper_bounds in enumerate(
                    ensemble_prediction.upper_bound
                ):
                    condition_id = ensemble_prediction.prediction_results[
                        0
                    ].condition_ids[i_cond]
                    write_array(ub_grp, condition_id, upper_bounds)
            elif isinstance(ensemble_prediction.upper_bound[0], float):
                f.create_dataset(
                    UPPER_BOUND, data=ensemble_prediction.upper_bound
                )

        # write summary statistics to h5 file
        for (
            summary_id,
            summary,
        ) in ensemble_prediction.prediction_summary.items():
            if summary is None:
                continue
            tmp_base_path = os.path.join(base, f'{SUMMARY}_{summary_id}')
            f.create_group(tmp_base_path)
            summary.write_to_h5(output_file, base_path=tmp_base_path)

        # write the single prediction results
        for i_result, result in enumerate(
            ensemble_prediction.prediction_results
        ):
            tmp_base_path = os.path.join(
                base, f'{PREDICTION_RESULTS}_{i_result}'
            )
            result.write_to_h5(output_file, base_path=tmp_base_path)


def get_prediction_dataset(
    ens: Union[Ensemble, EnsemblePrediction], prediction_index: int = 0
) -> np.ndarray:
    """
    Extract an array of prediction.

    Can be done from either an Ensemble object which contains a list of
    predictions of from an EnsemblePrediction object.

    Parameters
    ----------
    ens:
        Ensemble objects containing a set of parameter vectors and a set of
        predictions or EnsemblePrediction object containing only predictions

    prediction_index:
        index telling which prediction from the list should be analyzed

    Returns
    -------
    dataset:
        numpy array containing the ensemble predictions
    """
    if isinstance(ens, Ensemble):
        dataset = ens.predictions[prediction_index]
    elif isinstance(ens, EnsemblePrediction):
        ens.condense_to_arrays()
        dataset = ens.prediction_arrays[OUTPUT].transpose()
    else:
        raise Exception(
            'Need either an Ensemble object with predictions or '
            'an EnsemblePrediction object as input. Stopping.'
        )

    return dataset


def read_ensemble_prediction_from_h5(
    predictor: Union[Callable[[Sequence], PredictionResult], None],
    input_file: str,
):
    """Read an ensemble prediction from an HDF5 File."""
    # open file
    with h5py.File(input_file, 'r') as f:
        pred_res_list = []
        bounds = {}
        for key in f.keys():
            if key.startswith(SUMMARY):
                continue
            if key == PREDICTION_ID:
                prediction_id = f[key][()].decode()
                continue
            if key in {LOWER_BOUND, UPPER_BOUND}:
                if isinstance(f[key], h5py._hl.dataset.Dataset):
                    bounds[key] = f[key][:]
                    continue
                bounds[key] = [
                    f[f'{key}/{cond}'][()] for cond in f[key].keys()
                ]
                bounds[key] = np.array(bounds[key])
                continue
            x_names = list(decode_array(f[f'{key}/{X_NAMES}'][()]))
            condition_ids = list(decode_array(f[f'{key}/condition_ids'][()]))
            pred_cond_res_list = []
            for id, _ in enumerate(condition_ids):
                output = f[f'{key}/{id}/{OUTPUT}'][:]
                output_ids = tuple(
                    decode_array(f[f'{key}/{id}' f'/{OUTPUT_IDS}'][:])
                )
                timepoints = f[f'{key}/{id}/{TIMEPOINTS}'][:]
                try:
                    output_weight = f[f'{key}/{id}/{OUTPUT_WEIGHT}'][()]
                except KeyError:
                    output_weight = None
                try:
                    output_sigmay = f[f'{key}/{id}/{OUTPUT_SIGMAY}'][:]
                except KeyError:
                    output_sigmay = None
                pred_cond_res_list.append(
                    PredictionConditionResult(
                        timepoints=timepoints,
                        output_ids=output_ids,
                        output=output,
                        x_names=x_names,
                        output_weight=output_weight,
                        output_sigmay=output_sigmay,
                    )
                )
            pred_res_list.append(
                PredictionResult(
                    conditions=pred_cond_res_list, condition_ids=condition_ids
                )
            )
        return EnsemblePrediction(
            predictor=predictor,
            prediction_id=prediction_id,
            prediction_results=pred_res_list,
        )


def decode_array(array: np.ndarray) -> np.ndarray:
    """Decode array of bytes to string."""
    for i in range(len(array)):
        array[i] = array[i].decode()
    return array



"""Optimization result."""

import warnings
from collections import Counter
from copy import deepcopy
from typing import Sequence, Union

import numpy as np
import pandas as pd

from ..history import HistoryBase
from ..problem import Problem
from ..util import assign_clusters, delete_nan_inf

OptimizationResult = Union['OptimizerResult', 'OptimizeResult']


class OptimizerResult(dict):
    """
    The result of an optimizer run.

    Used as a standardized return value to map from the individual result
    objects returned by the employed optimizers to the format understood by
    pypesto.

    Can be used like a dict.

    Attributes
    ----------
    id:
        Id of the optimizer run. Usually the start index.
    x:
        The best found parameters.
    fval:
        The best found function value, `fun(x)`.
    grad:
        The gradient at `x`.
    hess:
        The Hessian at `x`.
    res:
        The residuals at `x`.
    sres:
        The residual sensitivities at `x`.
    n_fval
        Number of function evaluations.
    n_grad:
        Number of gradient evaluations.
    n_hess:
        Number of Hessian evaluations.
    n_res:
        Number of residuals evaluations.
    n_sres:
        Number of residual sensitivity evaluations.
    x0:
        The starting parameters.
    fval0:
        The starting function value, `fun(x0)`.
    history:
        Objective history.
    exitflag:
        The exitflag of the optimizer.
    time:
        Execution time.
    message: str
        Textual comment on the optimization result.
    optimizer: str
        The optimizer used for optimization.

    Notes
    -----
    Any field not supported by the optimizer is filled with None.
    """

    def __init__(
        self,
        id: str = None,
        x: np.ndarray = None,
        fval: float = None,
        grad: np.ndarray = None,
        hess: np.ndarray = None,
        res: np.ndarray = None,
        sres: np.ndarray = None,
        n_fval: int = None,
        n_grad: int = None,
        n_hess: int = None,
        n_res: int = None,
        n_sres: int = None,
        x0: np.ndarray = None,
        fval0: float = None,
        history: HistoryBase = None,
        exitflag: int = None,
        time: float = None,
        message: str = None,
        optimizer: str = None,
    ):
        super().__init__()
        self.id = id
        self.x: np.ndarray = np.array(x) if x is not None else None
        self.fval: float = fval
        self.grad: np.ndarray = np.array(grad) if grad is not None else None
        self.hess: np.ndarray = np.array(hess) if hess is not None else None
        self.res: np.ndarray = np.array(res) if res is not None else None
        self.sres: np.ndarray = np.array(sres) if sres is not None else None
        self.n_fval: int = n_fval
        self.n_grad: int = n_grad
        self.n_hess: int = n_hess
        self.n_res: int = n_res
        self.n_sres: int = n_sres
        self.x0: np.ndarray = np.array(x0) if x0 is not None else None
        self.fval0: float = fval0
        self.history: HistoryBase = history
        self.exitflag: int = exitflag
        self.time: float = time
        self.message: str = message
        self.optimizer = optimizer

    def __getattr__(self, key):
        try:
            return self[key]
        except KeyError:
            raise AttributeError(key)

    __setattr__ = dict.__setitem__
    __delattr__ = dict.__delitem__

    def summary(self):
        """Get summary of the object."""
        message = (
            "### Optimizer Result\n\n"
            f"* optimizer used: {self.optimizer}\n"
            f"* message: {self.message} \n"
            f"* number of evaluations: {self.n_fval}\n"
            f"* time taken to optimize: {self.time:0.3f}s\n"
            f"* startpoint: {self.x0}\n"
            f"* endpoint: {self.x}\n"
        )
        # add fval, gradient, hessian, res, sres if available
        if self.fval is not None:
            message += f"* final objective value: {self.fval}\n"
        if self.grad is not None:
            message += f"* final gradient value: {self.grad}\n"
        if self.hess is not None:
            message += f"* final hessian value: {self.hess}\n"
        if self.res is not None:
            message += f"* final residual value: {self.res}\n"
        if self.sres is not None:
            message += f"* final residual sensitivity: {self.sres}\n"

        return message

    def update_to_full(self, problem: Problem) -> None:
        """
        Update values to full vectors/matrices.

        Parameters
        ----------
        problem:
            problem which contains info about how to convert to full vectors
            or matrices
        """
        self.x = problem.get_full_vector(self.x, problem.x_fixed_vals)
        self.grad = problem.get_full_vector(self.grad)
        self.hess = problem.get_full_matrix(self.hess)
        self.x0 = problem.get_full_vector(self.x0, problem.x_fixed_vals)


class OptimizeResult:
    """Result of the :py:func:`pypesto.optimize.minimize` function."""

    def __init__(self):
        self.list = []

    def __deepcopy__(self, memo):
        other = OptimizeResult()
        other.list = deepcopy(self.list)
        return other

    def __getattr__(self, key):
        """Define `optimize_result.key`."""
        try:
            return [res[key] for res in self.list]
        except KeyError:
            raise AttributeError(key)

    def __getitem__(self, index):
        """Define `optimize_result[i]` to access the i-th result."""
        try:
            return self.list[index]
        except IndexError:
            raise IndexError(
                f"{index} out of range for optimize result of "
                f"length {len(self.list)}."
            )

    def __getstate__(self):
        # while we override __getattr__ as we do now, this is required to keep
        # instances pickle-able
        return vars(self)

    def __setstate__(self, state):
        # while we override __getattr__ as we do now, this is required to keep
        # instances pickle-able
        vars(self).update(state)

    def __len__(self):
        return len(self.list)

    def summary(self, disp_best: bool = True, disp_worst: bool = False) -> str:
        """
        Get summary of the object.

        Parameters
        ----------
        disp_best:
            Whether to display a detailed summary of the best run.
        disp_worst:
            Whether to display a detailed summary of the worst run.
        """
        if len(self) == 0:
            return "## Optimization Result \n\n*empty*\n"

        # perform clustering for better information
        clust, clustsize = assign_clusters(delete_nan_inf(self.fval)[1])

        # aggregate exit messages
        message_counts_df = pd.DataFrame(
            Counter(self.message).most_common(), columns=["Message", "Count"]
        )
        counter_message = message_counts_df[["Count", "Message"]].to_markdown(
            index=False
        )
        counter_message = "  " + counter_message.replace("\n", "\n  ")

        times_message = (
            f'\t* Mean execution time: {np.mean(self.time):0.3f}s\n'
            f'\t* Maximum execution time: {np.max(self.time):0.3f}s,'
            f'\tid={self[np.argmax(self.time)].id}\n'
            f'\t* Minimum execution time: {np.min(self.time):0.3f}s,\t'
            f'id={self[np.argmin(self.time)].id}'
        )

        # special handling in case there are only non-finite fvals
        num_best_value = int(clustsize[0]) if len(clustsize) else len(self)
        num_plateaus = (
            (1 + max(clust) - sum(clustsize == 1)) if len(clustsize) else 0
        )

        summary = (
            "## Optimization Result \n\n"
            f"* number of starts: {len(self)} \n"
            f"* best value: {self[0]['fval']}, id={self[0]['id']}\n"
            f"* worst value: {self[-1]['fval']}, id={self[-1]['id']}\n"
            f"* number of non-finite values: "
            f"{np.logical_not(np.isfinite(self.fval)).sum()}\n\n"
            f"* execution time summary:\n{times_message}\n"
            f"* summary of optimizer messages:\n\n{counter_message}\n\n"
            f"* best value found (approximately) {num_best_value} time(s)\n"
            f"* number of plateaus found: {num_plateaus}\n"
        )
        if disp_best:
            summary += f"\nA summary of the best run:\n\n{self[0].summary()}"
        if disp_worst:
            summary += f"\nA summary of the worst run:\n\n{self[-1].summary()}"
        return summary

    def append(
        self,
        optimize_result: OptimizationResult,
        sort: bool = True,
        prefix: str = '',
    ):
        """
        Append an OptimizerResult or an OptimizeResult to the result object.

        Parameters
        ----------
        optimize_result:
            The result of one or more (local) optimizer run.
        sort:
            Boolean used so we only sort once when appending an
            optimize_result.
        prefix:
            The IDs for all appended results will be prefixed with this.
        """
        current_ids = set(self.id)
        if isinstance(optimize_result, OptimizeResult):
            new_ids = {
                prefix + identifier
                for identifier in optimize_result.id
                if identifier is not None
            }
            if not current_ids.isdisjoint(new_ids):
                raise ValueError(
                    "Some id's you want to merge coincide with "
                    "the existing id's. Please use an "
                    "appropriate prefix such as 'run_2_'."
                )
            for optimizer_result in optimize_result.list:
                self.append(optimizer_result, sort=False, prefix=prefix)
        elif isinstance(optimize_result, OptimizerResult):
            # if id is None, append without checking for duplicate ids
            if optimize_result.id is None:
                self.list.append(optimize_result)
            else:
                new_id = prefix + optimize_result.id
                if new_id in current_ids:
                    raise ValueError(
                        "The id you want to merge coincides with "
                        "the existing id's. Please use an "
                        "appropriate prefix such as 'run_2_'."
                    )
                optimize_result.id = new_id
                self.list.append(optimize_result)
        if sort:
            self.sort()

    def sort(self):
        """Sort the optimizer results by function value fval (ascending)."""

        def get_fval(res):
            return res.fval if not np.isnan(res.fval) else np.inf

        self.list = sorted(self.list, key=get_fval)

    def as_dataframe(self, keys=None) -> pd.DataFrame:
        """
        Get as pandas DataFrame.

        If keys is a list, return only the specified values, otherwise all.
        """
        lst = self.as_list(keys)

        df = pd.DataFrame(lst)

        return df

    def as_list(self, keys=None) -> Sequence:
        """
        Get as list.

        If keys is a list, return only the specified values.

        Parameters
        ----------
        keys: list(str), optional
            Labels of the field to extract.
        """
        lst = self.list

        if keys is not None:
            lst = [{key: res[key] for key in keys} for res in lst]

        return lst

    def get_for_key(self, key) -> list:
        """Extract the list of values for the specified key as a list."""
        warnings.warn(
            "get_for_key() is deprecated in favour of "
            "optimize_result.key and will be removed in future "
            "releases."
        )
        return [res[key] for res in self.list]



from typing import List

import numpy as np


class ReferencePoint(dict):
    """
    Reference point for plotting.

    Should contain a parameter value and an objective function value,
    may also contain a color and a legend.

    Can be used like a dict.

    Attributes
    ----------
    x: ndarray
        Reference parameters.
    fval: float
        Function value, fun(x), for reference parameters.
    color: RGBA, optional
        Color which should be used for reference point.
    auto_color: boolean
        flag indicating whether color for this reference point should be
        assigned automatically or whether it was assigned by user
    legend: str
        legend text for reference point
    """

    def __init__(
        self, reference=None, x=None, fval=None, color=None, legend=None
    ):
        super().__init__()

        if (reference is not None) and ((x is not None) or (fval is not None)):
            raise ValueError(
                "Please specify either an argument for reference "
                "or for x and fval, but not both."
            )

        # assign legend, may be None
        self.legend = legend
        if isinstance(reference, dict) or isinstance(
            reference, ReferencePoint
        ):
            # Handle case of dict or ReferencePoint
            self.x = np.array(reference["x"])
            self.fval = reference["fval"]
            if "color" in reference.keys():
                self.color = reference["color"]
                if "auto_color" in reference.keys():
                    self.auto_color = reference["auto_color"]
                else:
                    self.auto_color = False
            else:
                self.color = None
                self.auto_color = True
            if "legend" in reference.keys():
                self.legend = reference["legend"]
            else:
                self.legend = None
        elif isinstance(reference, tuple):
            # Handle case of tuple
            self.x = np.array(reference[0])
            self.fval = reference[1]
            if len(reference) > 2:
                self.color = reference[2]
                self.auto_color = False
            else:
                self.color = None
                self.auto_color = True
        if reference is None:
            if x is not None:
                self.x = np.array(x)
            else:
                raise ValueError(
                    'Parameter vector x not passed, but is a '
                    'mandatory input when creating a reference '
                    'point. Stopping.'
                )
            if fval is not None:
                self.fval = fval
            else:
                raise ValueError(
                    'Objective value fval not passed, but is a '
                    'mandatory input when creating a reference '
                    'point. Stopping.'
                )
            if color is not None:
                self.color = color
                self.auto_color = False
            else:
                self.color = None
                self.auto_color = True

    def __getattr__(self, key):
        """Allow usage of keys as attributes."""
        try:
            return self[key]
        except KeyError:
            raise AttributeError(key)

    __setattr__ = dict.__setitem__
    __delattr__ = dict.__delitem__


def assign_colors(ref):
    """
    Assign colors to reference points, depending on user settings.

    Parameters
    ----------
    ref: list of ReferencePoint
        Reference points, which need to get their color property filled

    Returns
    -------
    ref: list of ReferencePoint
        Reference points, which got their color property filled
    """
    # loop over reference points
    auto_color_count = 0
    for i_ref in ref:
        if i_ref['auto_color']:
            auto_color_count += 1

    auto_colors = [
        [0.0, 0.5 * (1.0 + i_auto / auto_color_count), 0.0, 0.9]
        for i_auto in range(auto_color_count)
    ]

    # loop over reference points and assign auto_colors
    auto_color_count = 0
    for i_num, i_ref in enumerate(ref):
        if i_ref['auto_color']:
            i_ref['color'] = auto_colors[i_num]
            auto_color_count += 1

    return ref


def create_references(
    references=None, x=None, fval=None, color=None, legend=None
) -> List[ReferencePoint]:
    """
    Create a list of reference point objects from user inputs.

    Parameters
    ----------
    references: ReferencePoint or dict or list, optional
        Will be converted into a list of RefPoints
    x: ndarray, optional
        Parameter vector which should be used for reference point
    fval: float, optional
        Objective function value which should be used for reference point
    color: RGBA, optional
        Color which should be used for reference point.
    legend: str
        legend text for reference point

    Returns
    -------
    colors: list of RGBA
        One for each element in 'vals'.
    """
    # parse input (reference)
    ref = []
    if references is not None:
        if isinstance(references, list):
            for reference in references:
                ref.append(ReferencePoint(reference))
        else:
            ref.append(ReferencePoint(references))

    # parse input (x and fval)
    if (x is not None) and (fval is not None):
        ref.append(ReferencePoint(x=x, fval=fval, color=color, legend=legend))

    # assign colors for reference points which have no user-specified colors
    return assign_colors(ref)
